//
// $Id: generic-ast.h 2 2007-12-30 09:48:00Z hat $
//
#ifndef H_GENERIC_AST
#define H_GENERIC_AST

#include "types.h"
#include "textstring.h"
#include "byte-sequence.h"
#include <ostream>

using std::ostream;

/** @file: generic-ast.h
 * Generic AST is the generic infra structure for storing hierarchical
 * key-value data without interpreting it.
 */

/* {{{ IdentifierLine, IdentifierSet
 */

/** Class containing a single identifier and a line number
 */
class IdentifierLine
{
public:
	IdentifierLine();
	IdentifierLine(const string& sIden, int iLine);
	IdentifierLine(const IdentifierLine&);
	IdentifierLine& operator=(const IdentifierLine&);
	~IdentifierLine();

	string get_iden() const;
	int get_line() const;
protected:
	string sIden;
	int iLine;
};

inline bool operator<(const IdentifierLine& cil1, const IdentifierLine& cil2)
{
	string s1 = cil1.get_iden();
	string s2 = cil2.get_iden();

	if(s1 < s2) return true;
	if(s2 < s1) return false;
	// s1 == s2
	return cil1.get_line() < cil2.get_line();
}

/** Set of identifiers + line numbers
 */
typedef set<IdentifierLine> IdentifierSet;

/* }}}
 * {{{ Generic fields, Node base class
 */

/** Base class for generic field in the generic AST
 * */
class BaseField
{
public:
    BaseField(int line=0);
    virtual ~BaseField();

    void set_line(int lin);
    int get_line() const;
protected:
    int iLine;
};

/** Base class for a key field in the generic AST
 * */
class Field: public BaseField
{
public:
	Field(const string&, int line);
	virtual ~Field();
//protected:
	string sName;	//< Name/key of the field
};

/** Base class for nodes
 */
class GenericNode: public BaseField
{
public:
	GenericNode();
	~GenericNode();

	void add_field(BaseField *pBF);
	void set_name(const string& sN);
	void set_label(const string& sL, int line);
// protected:
	list<BaseField*> cFields;	//< List of fields in this node
	string sName,  //< Name of the generic node (used as key while decoding)
	       sLabel; //< Name of the label of this generic node
};

/** A generic field consisting just of a single key name (no value)
 */
class FieldKey: public Field
{
public:
	FieldKey(const string&, int line);
	~FieldKey();
};

/** A generic field wth a key and an identifier
 */
class FieldIdenValue: public Field
{
public:
	FieldIdenValue(const string& key, int key_line,
	           const string& iden_val, int iden_line);
	~FieldIdenValue();

//protected:
	string sIdenValue;	//< Value identifier of the field
	int iIdenValueLine;	//< Line of the identifier value
};

/** A generic field with a set of identifiers
 */
class FieldIdentifierSet: public Field
{
public:
	FieldIdentifierSet(const string& key, int key_line,
	                   const IdentifierSet& iden_set);
	~FieldIdentifierSet();

//protected:
	IdentifierSet cIdentifiers;
};

/** A generic field with an unsigned value
 */
class FieldUnsigned: public Field
{
public:
	FieldUnsigned(const string& key, int key_line, UINT32 val, int line);
	~FieldUnsigned();

//protected:
	UINT32 value;
	int line;
};

/** A generic field with a signed value
 */
class FieldSigned: public Field
{
public:
	FieldSigned(const string& key, int key_line, INT32 val, int line);
	~FieldSigned();

//protected:
	INT32 value;
	int line;
};

/** A generic field with a TextString
 */
class FieldString: public Field
{
public:
	FieldString(const string& key, int key_line, const TextString& cTxt);
	~FieldString();

//protected:
	TextString cText;
};

/** A generic field with a sequence of bytes (a ByteSequence)
 */
class FieldBytes: public Field
{
public:
	FieldBytes(const string& key, int key_line, const ByteSequence& cBytes);
	~FieldBytes();

//protected:
	ByteSequence cBytes;
};

/** A generic field with a list of labels (to sprite-sets)
 */
class LabelLine
{
public:
    LabelLine();
    LabelLine(const string& sLabel, int line);
    LabelLine(const LabelLine& ll);
    LabelLine& operator=(const LabelLine&);
    ~LabelLine();
//protected:
    string sLabel;
    int line;
};

typedef list<LabelLine> LabelLineList;

class FieldLabels: public Field
{
public:
    FieldLabels(const string& key, int key_line,
                                           const list<LabelLine>& cLabels);
    ~FieldLabels();

//protected:
    list<LabelLine> cLabels;
};

/* }}}
 */

#endif
