//
// $Id: node.cc 2 2007-12-30 09:48:00Z hat $
//
#include "generic-ast.h"
#include "node.h"
#include "errors.h"
#include "codehelper.h"
#include "table.h"
#include <cassert>

#include <iostream>

/* {{{ enum ESize: size_name(), unsigned_in_range(), signed_in_range() */

string size_name(ESize e)
{
    if(e == ES_Byte) return string("byte");
    if(e == ES_Word) return string("word");
    if(e == ES_Double) return string("double");
    assert(0);
}

UINT32 clip_unsigned(UINT32 val, ESize e)
{
    if(e == ES_Byte) return val & 0xff;
    if(e == ES_Word) return val & 0xffff;
    return val;
}

INT32 clip_signed(INT32 val, ESize e)
{
    if(e == ES_Byte) {
        if(val < -128) return -128;
        if(val > 127) return 127;
        return val;
    }
    if(e == ES_Word) {
        if(val < -32768) return -32768;
        if(val > 32767) return 32767;
        return val;
    }
    return val;
}

bool unsigned_in_range(UINT32 val, ESize e)
{
    if(e == ES_Byte) return val <= 0xff;
    if(e == ES_Word) return val <= 0xffff;
    return true;
}

bool signed_in_range(INT32 val, ESize e)
{
    if(e == ES_Byte) return val >= -128 && val <= 127;
    if(e == ES_Word) return val >= -32768 && val <= 32767;
    return true;
}

void write_sized_data(CodeHelper& ch, UINT32 val, ESize es)
{
    if(es == ES_Byte) { ch.add_byte(val); return; }
    if(es == ES_Word) { ch.add_word(val); return; }
    ch.add_double(val);
}

/* }}} */

// {{{ static BaseField* check_assign_field(int line, BaseField*pOld, BaseField *pNew)
/** Check whether the argument \a pOld may be overwritten by argument \a pNew.
 * If needed report a warning. Also delete the old data.
 * @return: The new data
 */
static BaseField* check_assign_field(int line, BaseField*pOld, BaseField *pNew)
{
    assert(pNew);

    if(pOld) {
        char buffer[256];
        sprintf(buffer, "Overwriting field at line %d by field at line %d",
                pOld->get_line(), pNew->get_line());

        report_warning(line, buffer);
        delete pOld;
    }
    return pNew;
}
// }}}
// {{{ static void add_collections(MainNode *pMain, CodeHelper& ch,
/** Add word-indices of sprite collections to the CodeHelper */
static void add_collections(MainNode *pMain, CodeHelper& ch,
                            const LabelLineList& cLabels)
{
    list<LabelLine>::const_iterator iter;
    for(iter = cLabels.begin(); iter != cLabels.end(); ++iter) {
        int iCollection = pMain->findCollection((*iter).sLabel);
        if(iCollection < 0) {
            report_error((*iter).line, string("Collection \"") +
                         (*iter).sLabel + string("\" not available (here)"));
            iCollection = 0;
        }
        ch.add_word(iCollection);
    }
}
// }}}
// {{{ static bool add_cid(MainNode *pMain, CodeHelper& ch, BaseField *pCID)
/** Add a cid word to the code helper for the label refered to by \a pCID.
 * @param pMain: \c MainNode object (contains label -> cid map)
 * @param ch: \c CodeHelper (to write word to)
 * @param pCID: Field value containing the label
 * @return: Boolean indicating success. If \c false returned, give an error
 *          only, a word has already been written to the code helper.
 */
static bool add_cid(MainNode *pMain, CodeHelper& ch, BaseField *pCID)
{
    FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pCID);
    if(pU) {
        if(pU->value > 0xffff) {
            report_error(pU->get_line(), "Value too large");
            ch.add_word(0xffff);
        } else {
            ch.add_word(pU->value);
        }
        return true;
    }

    FieldIdenValue *pI = dynamic_cast<FieldIdenValue*>(pCID);
    if(pI) {
        int labelnum;
        labelnum = pMain->findGraphics(pI->sIdenValue);
        if(labelnum >= 0) {
            ch.add_word(labelnum);
        } else {
            report_error(pI->get_line(), string("Label '") + pI->sIdenValue +
                                         string("' not available (here)"));
            ch.add_word(0);
        }
        return true;
    }

    ch.add_word(0); // fill missing word, reurn error indication
    return false;
}
// }}}

/* {{{ Node
 */
Node::Node(const string& sN, const string& sL, int lin)
	: BaseField(lin), iSpriteNumber(0), sName(sN), sLabel(sL) { }
Node::~Node() { }

string Node::get_name() { return sName; }
string Node::get_label() const { return sLabel; }
int Node::get_sprite_number() { return iSpriteNumber; }

/** Set the sprite number of the first sprite to sprnum, return next
 * free sprite number
 */
int Node::set_sprite_number(int sprnum)
{
    iSpriteNumber = sprnum;
    return sprnum + 1;
}

void Node::generate_nfo(MainNode *pMain, FILE *fp)
{
    std::cerr << "generate_nfo() not implemented for node " << sName
             << " at line " << get_line() << "\n";
}

/* {{{ bool Node::unsigned_idenvalue()
 */
/** Decode an unsigned number/identifier. Return decoded data into \a pData.
 * Fields of \a pData are set to 0 initially.
 *
 * @param sFieldName: Name of the field being decoded
 * @param pField: Pointer to the field
 * @param sTableName: Name of the table to search for identifiers, empty means
 *      'no identifiers available'.
 * @param eSize: Expected size of the value
 * @param pData: Pointer to the returned data
 * @return: Boolean indicating success
 */
bool Node::unsigned_idenvalue(const string& sFieldName, BaseField *pField,
                              const string& sTableName, ESize eSize,
                              TableEntry* pData)
{
    assert(pData); pData->reset();

    if(!pField) {
        report_error(get_line(), string("Missing field ") + sFieldName);
        return false;
    }

    FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pField);
    if(pU) {
        if(!unsigned_in_range(pU->value, eSize)) {
            report_error(pU->line, string("Value to large to fit in a ") +
                                   size_name(eSize));
        }
        pData->set_value(clip_unsigned(pU->value, eSize));
        return true;
    }

    if(sTableName.size() > 0) {
        FieldIdenValue *pFI = dynamic_cast<FieldIdenValue*>(pField);
        if(pFI) {
            /* We do not check the size of the value associated with the
             * identifier, since we trust those tables
             */
            const Table *pTable = cTables.get_table(sTableName);
            const TableEntry *pEntry = pTable->get_entry(pFI->sIdenValue);
            if(pEntry) {
                if(pEntry->has_name()) pData->set_name(pEntry->get_name());
                if(pEntry->has_value()) pData->set_value(pEntry->get_value());
                if(pEntry->has_size()) pData->set_size(pEntry->get_size());
                if(pEntry->has_type()) pData->set_type(pEntry->get_type());
                return true;
            }

            report_error(pFI->iIdenValueLine,
                         string("Unrecognized identifier '") +
                         pFI->sIdenValue + string("' in table '") +
                         sTableName + string("'"));
            return false;
        }
    }

    /* We do not check the default value, since we trust that as well */
    report_error(pField->get_line(), "Unrecognized field value");
    return false;
}

/* }}}
 * {{{ Node::decode_unsigned()
 */
/** Decode an unsigned number
 * @param sFieldName: Name of the field being decoded
 * @param pField: Pointer to the field
 * @param default_val: Default value to use
 * @return: Data value
 */
UINT32 Node::decode_unsigned(const string& sFieldName,
                           BaseField *pField, ESize eSize, UINT32 default_val)
{
    TableEntry te;
    if(!unsigned_idenvalue(sFieldName, pField, string(""), eSize, &te))
        return default_val;

    return te.get_value();
}

/* }}}
 * {{{ Node::decode_signed()
 */
/** Decode a signed number
 * @param sFieldName: Name of the field being decoded
 * @param pField: Pointer to the field
 * @param default_val: Default value to use
 * @return: Data value
 */
INT32 Node::decode_signed(const string& sFieldName,
                           BaseField *pField, ESize eSize, INT32 default_val)
{
    if(!pField) {
        report_error(get_line(), string("Missing field ") + sFieldName);
        return default_val;
    }

    FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pField);
    if(pU) {
        assert(pU->value <= 0x7fffffff);
        return clip_signed((INT32)(pU->value), eSize);
    }

    FieldSigned *pS = dynamic_cast<FieldSigned*>(pField);
    if(pS) {
        return clip_signed(pS->value, eSize);
    }

    report_error(pField->get_line(), "Unrecognized field value");
    return default_val;
}

/* }}}
 * {{{ bool Node::bitset_idenvalue()
 */
/** Decode an unsigned number/identifier-set
 * @param sFieldName: Name of the field being decoded
 * @param pField: Pointer to the field
 * @param sTableName: Name of the table to search for identifiers, empty means
 *      'no identifiers available'.
 * @param eSize: Expected size of the value
 * @param default_val: Default value to use
 * @return: Data value (byte)
 */
bool Node::bitset_idenvalue(const string& sFieldName, BaseField *pField,
                            const string& sTableName, ESize eSize,
                            TableEntry *pData)
{
    UINT32 value;

    assert(pData); pData->reset();

    if(!pField) {
        report_error(get_line(), string("Missing field ") + sFieldName);
        return false;
    }

    FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pField);
    if(pU) {
        if(!unsigned_in_range(pU->value, eSize)) {
            report_error(pU->line, string("Value to large to fit in a ") +
                                   size_name(eSize));
        }
        pData->set_value(clip_unsigned(pU->value, eSize));
        return true;
    }

    if(sTableName.size() > 0) {
        const Table *pTable = cTables.get_table(sTableName);
        FieldIdenValue *pFI = dynamic_cast<FieldIdenValue*>(pField);
        if(pFI) {
            const TableEntry *pEntry = pTable->get_entry(pFI->sIdenValue);
            if(pEntry) {
                pData->set_value(pEntry->get_value());
                return true;
            }

            report_error(pFI->iIdenValueLine, "Unrecognized identifier");
            return false;
        }

        FieldIdentifierSet *pFIS = dynamic_cast<FieldIdentifierSet*>(pField);
        if(pFIS) {
            value = 0;
            set<IdentifierLine>::iterator iter;
            for(iter = pFIS->cIdentifiers.begin();
                                    iter != pFIS->cIdentifiers.end(); ++iter) {
                const TableEntry *pEntry;
                pEntry = pTable->get_entry((*iter).get_iden());
                if(pEntry) {
                    value |= pEntry->get_value();
                } else {
                    report_error((*iter).get_line(), "Unrecognized identifier");
                }
            }
            pData->set_value(value);
            return true;
        }
    }

    report_error(pField->get_line(), "Unrecognized field value");
    return false;
}

/* }}}
 * {{{ Node::generate_string()
 */
/** Generate a null-terminated string in the output
 * @param node_line: Line number of the node
 * @param sName: Name of the field
 * @param pBF: Data supplied by the user
 * @param ch: Output helper object
 */
void Node::generate_string(int node_line, const string& sName,
                           BaseField *pBF, CodeHelper& ch)
{
    if(!pBF) {
        report_error(node_line, string("Missing field ") + sName);
        char *pTxt = "Unknown";
        while(pTxt) {
            ch.add_byte(*pTxt); ++pTxt;
        }
        ch.add_byte(0); return;
    }
    FieldString *pFS = dynamic_cast<FieldString*>(pBF);
    if(pFS) {
        pFS->cText.add_text(ch, 0);
        ch.add_byte(0); return;
    }
    FieldBytes *pB = dynamic_cast<FieldBytes*>(pBF);
    if(pB) {
        pB->cBytes.copy_to_codehelper(ch); return;
    }

    report_error(pBF->get_line(), "Unrecognized field value");
    char *pTxt = "Unknown";
    while(pTxt) {
        ch.add_byte(*pTxt); ++pTxt;
    }
    ch.add_byte(0); return;
}

/* }}}
 * {{{ UINT32 Node::decode_language(MainNode *pMain)
 */
/** Decode language field to byte value
 */
UINT32 Node::decode_language(BaseField *pLanguage, MainNode *pMain)
{
    TableEntry lang_entry;

    /* language depends on the grlid version */
    int version = pMain->get_grlid_version();
    if(version > 6) {
        if(!unsigned_idenvalue(string("language"), pLanguage,
                               string("language_version7_table"), ES_Byte,
                               &lang_entry))
            lang_entry.set_value(0x00); // american
    } else {
        if(!bitset_idenvalue(string("language"), pLanguage,
                             string("language_version6_table"), ES_Byte,
                             &lang_entry))
            lang_entry.set_value(0x01); // american
    }
    return lang_entry.get_value();
}
/*
 * }}}
 * }}}  End Node
 *
 * {{{ CondJumpNode
 */
CondJumpNode::CondJumpNode(const string& sName, GenericNode *pGN)
        : Node(sName, pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    init_cond = (sName == "init-cond");
    pVariable = 0; pTest = 0; pValue = 0; pVarsize = 0; pDestination = 0;

    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "variable") {
            pVariable = check_assign_field(pGN->get_line(),
                                               pVariable, *iter);
            continue;
        }
        if(pF && pF->sName == "test") {
            pTest = check_assign_field(pGN->get_line(), pTest, *iter);
            continue;
        }
        if(pF && pF->sName == "value") {
            pValue = check_assign_field(pGN->get_line(), pValue, *iter);
            continue;
        }
        if(pF && pF->sName == "var-size") {
            pVarsize = check_assign_field(pGN->get_line(), pVarsize, *iter);
            continue;
        }
        if(pF && pF->sName == "destination") {
            pDestination = check_assign_field(pGN->get_line(),
                                              pDestination, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}
CondJumpNode::~CondJumpNode()
{
    if(pVariable) delete pVariable;
    if(pTest) delete pTest;
    if(pValue) delete pValue;
    if(pVarsize) delete pVarsize;
    if(pDestination) delete pDestination;
}

void CondJumpNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    UINT32 value, size;
    int offset;

    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(init_cond ? 9 : 7); // action 9/7

    /* decode condition variable */
    TableEntry var_entry;
    if(!unsigned_idenvalue(string("variable"), pVariable,
                           string("condition_variable_table"),
                           ES_Byte, &var_entry))
        var_entry.set_value(0x81); // 0x81=year

    /* decode condition test */
    TableEntry test_entry;
    if(!unsigned_idenvalue(string("test"), pTest,
                                string("condition_test_table"),
                                ES_Byte, &test_entry))
        test_entry.set_value(2);    // 2=equal

    /* check correctness of test/type (if available) */
    if(var_entry.has_type() && test_entry.has_type()
                            && !(var_entry.get_type() & test_entry.get_type()))
        report_error(get_line(), "Variable and test do not match");

    pseudo.add_byte(var_entry.get_value());

    /* compute/get varsize */
    if(var_entry.has_size()) {
        /* variable has size indication */
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pVarsize);
        if(pU) {
            if(pU->value != var_entry.get_size())
                report_warning(pU->get_line(),
                             "Specified size does not match with variable");
            size = pU->value & 0xff;
        } else
            size = var_entry.get_size();
    } else {
        /* variable has no size indication */
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pVarsize);
        if(pU) {
            if(pU->value != 1 && pU->value != 4)
                report_error(pU->get_line(), "Unrecognized var-size");
            size = pU->value & 0xff;
        } else {
            report_error(get_line(), "Missing var-size field");
            size = 1;   // default value
        }
    }
    pseudo.add_byte(size);  // write size
    pseudo.add_byte(test_entry.get_value());  // write test

    /* decode value */
    if(size == 1 || size == 0) {
        TableEntry val_entry;
        if(!unsigned_idenvalue(string("value"), pValue,
                               string("condition_value_table"), ES_Byte,
                               &val_entry))
            val_entry.set_value(0);

        /* check type of test wrt the value */
        if(val_entry.has_type() && var_entry.has_type() &&
                            !(var_entry.get_type() & val_entry.get_type()))
            report_error(get_line(), "Variable and value do not match");
        if(val_entry.has_type() && test_entry.has_type() &&
                            !(test_entry.get_type() & val_entry.get_type()))
            report_error(get_line(), "Test and value do not match");

        pseudo.add_byte(val_entry.get_value());
    } else {
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pValue);
        if(pU) {
            value = pU->value;
        } else {
            report_error(get_line(), "Unrecognized or missing value field");
            value = 0;
        }

        pseudo.add_double(value);
    }

    /* destination */
    FieldIdenValue *pI = dynamic_cast<FieldIdenValue*>(pDestination);
    if(pI) {
        if(pI->sIdenValue == string("exit")) {
            pseudo.add_byte(0);
        } else {
            Node *destNode = pMain->get_destination(pI->sIdenValue);
            if(!destNode) {
                report_error(pI->get_line(), "Unknown destination");
                offset = 0;
            } else {
                offset = destNode->get_sprite_number() - iSpriteNumber;
                if(offset < 0) {
                    report_error(pI->get_line(), "Should not jump back");
                    offset &= 0xff;
                }
                offset -= 1;
                if(offset == 0) {
                    report_error(pI->get_line(),
                                 "Should not jump to next sprite");
                }
                if(offset > 255) {
                    report_error(pI->get_line(), "Destination too far away");
                    offset &= 0xff;
                }
            }
            pseudo.add_byte(offset);
        }
    } else {
        report_error(get_line(), "Missing destination field");
        pseudo.add_byte(0);
    }

    pseudo.output(fp);
}

/* }}}
 * {{{ MessageNode
 */
MessageNode::MessageNode(GenericNode *pGN)
	: Node(string("message"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pSeverity = 0; pData = 0; pText = 0; pLanguage = 0;
    pParnum1 = 0; pParnum2 = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "severity") {
            pSeverity = check_assign_field(pGN->get_line(), pSeverity, *iter);
            continue;
        }
        if(pF && pF->sName == "data") {
            pData = check_assign_field(pGN->get_line(), pData, *iter);
            continue;
        }
        if(pF && pF->sName == "text") {
            pText = check_assign_field(pGN->get_line(), pText, *iter);
            continue;
        }
        if(pF && pF->sName == "language") {
            pLanguage = check_assign_field(pGN->get_line(), pLanguage, *iter);
            continue;
        }
        if(pF && pF->sName == "parnum1") {
            pParnum1 = check_assign_field(pGN->get_line(), pParnum1, *iter);
            continue;
        }
        if(pF && pF->sName == "parnum2") {
            pParnum2 = check_assign_field(pGN->get_line(), pParnum2, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

MessageNode::~MessageNode()
{
    if(pSeverity) delete pSeverity;
    if(pText) delete pText;
    if(pData) delete pData;
    if(pLanguage) delete pLanguage;
    if(pParnum1) delete pParnum1;
    if(pParnum2) delete pParnum2;
}

static const DataValue file_data_parm_table[] =
{
    {"file", 0x80}, {"data", 0x80}, {"parm", 0x7b}, {0,0}
};

void MessageNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    UINT32 value;
    FieldString *pFieldText;
    int idx;

    CodeHelper pseudo(iSpriteNumber);
    pseudo.add_byte(0xb);   // action code

    /* get severity byte */
    TableEntry val_entry;
    if(!unsigned_idenvalue(string("severity"), pSeverity,
                           string("severity_table"), ES_Byte,
                           &val_entry))
        val_entry.set_value(0);    // 0=severity-notice

    pseudo.add_byte(val_entry.get_value());

    /* language of the message, high bit is always clear */
    pseudo.add_byte( decode_language(pLanguage, pMain) );

    /* text message */
    pFieldText = dynamic_cast<FieldString*>(pText);
    if(pFieldText) {
        value = 0xff;
        if(value == 0xff && pFieldText->cText == TextString("%{file} "
                                "requires at least TTDPatch version %{data}"))
            value = 0; 
    
	if(value == 0xff && pFieldText->cText == TextString("%{file} is for "
                                "the %{data} version of TTD."))
            value = 1;

        if(value == 0xff && pFieldText->cText == TextString("%{file} is "
                                "designed to be used with %{data}"))
            value = 2;


	if(value == 0xff && pFieldText->cText == TextString("Invalid "
                        "parameter for %{file}: parameter %{data} (%{parm})"))
            value = 3;

	if(value == 0xff && pFieldText->cText == TextString("%{file} must be "
                                "loaded before %{data}."))
            value = 4;

	if(value == 0xff && pFieldText->cText == TextString("%{file} must be "
                                "loaded after %{data}."))
            value = 5;

        pseudo.add_byte(value);
        if(value == 0xff) { // custom message
            pFieldText->cText.add_text(pseudo, file_data_parm_table);
            pseudo.add_byte(0); // terminating \0
        }
    } else {
        //pFieldText = 0;
        report_error(get_line(), "Missing or incorrect text field");
        pseudo.add_byte(0);
    }

    /* data field */
    FieldString *pS = dynamic_cast<FieldString*>(pData);
    if(pS) {
        pS->cText.add_text(pseudo, 0);
        pseudo.add_byte(0); // terminating \0
    } else {
        report_error(get_line(), "Missing or incorrect data field");
        pseudo.add_byte(0);
    }

    /* parnum1 & 2 */
    if(pFieldText) {
        idx = pFieldText->cText.find_esc_text(string("parnum"), 0);
        if(idx >= 0) {
            FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pParnum1);
            if(pU) {
                if(pU->value >= 0 && pU->value <= 9)
                    pseudo.add_byte('0' + pU->value);
                else {
                    report_error(pU->get_line(), "Incorrect parnum1 value");
                    pseudo.add_byte(0);
                }
            } else {
                report_error(get_line(), "Missing parnum1 field");
            }
        }
        /* parnum2 */
        idx = pFieldText->cText.find_esc_text(string("parnum"), idx + 1);
        if(idx >= 0) {
            FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pParnum2);
            if(pU) {
                if(pU->value >= 0 && pU->value <= 9)
                    pseudo.add_byte('0' + pU->value);
                else {
                    report_error(pU->get_line(), "Incorrect parnum2 value");
                    pseudo.add_byte(0);
                }
            } else {
                report_error(get_line(), "Missing parnum2 field");
            }
        }
    }

    pseudo.output(fp);
}

/* }}}
 * {{{ GrlIdNode
 */
GrlIdNode::GrlIdNode(GenericNode *pGN)
        : Node(string("grl-id"), pGN->sLabel, pGN->get_line()), version(-1)
{
    list<BaseField*>::iterator iter;

    pVersion = 0; pId = 0; pName = 0; pDescription = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "version") {
            pVersion = check_assign_field(pGN->get_line(), pVersion, *iter);
            continue;
        }
        if(pF && pF->sName == "id") {
            pId = check_assign_field(pGN->get_line(), pId, *iter);
            continue;
        }
        if(pF && pF->sName == "name") {
            pName = check_assign_field(pGN->get_line(), pName, *iter);
            continue;
        }
        if(pF && pF->sName == "description") {
            pDescription = check_assign_field(pGN->get_line(),
                                              pDescription, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

GrlIdNode::~GrlIdNode()
{
    if(pVersion) delete pVersion;
    if(pId) delete pId;
    if(pName) delete pName;
    if(pDescription) delete pDescription;
}

int GrlIdNode::get_version()
{
    if(version != -1)
        return version;

    TableEntry version_entry;
    if(!unsigned_idenvalue(string("version"), pVersion,
                           string(""), ES_Byte, &version_entry))
        version_entry.set_value(6); // default version 6

    int val = version_entry.get_value();
    if(val == 0 || val == 1 || val == 4 || val == 5 || val == 6 || val == 7)
        version = val;
    else {
        report_error(get_line(), "Incorrect version number found");
        version = 6;
    }
    return version;
}

void GrlIdNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(8); // action 8
    pseudo.add_byte(get_version());
    generate_id(pseudo);
    generate_string(get_line(), string("name"), pName, pseudo);
    generate_string(get_line(), string("description"), pDescription, pseudo);
    pseudo.output(fp);
}

void GrlIdNode::generate_id(CodeHelper& ch)
{
    if(!pId) {
        report_error(get_line(), "Missing ID field");
        ch.add_double(0ul); return;
    }
    FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pId);
    if(pU) {
        ch.add_double(pU->value); return;
    }

    FieldBytes *pB = dynamic_cast<FieldBytes*>(pId);
    if(pB) {
        int length = pB->cBytes.byte_length();
        if(length < 4) {
            report_error(pB->get_line(), "Not enough bytes in ID (4 expected)");
            pB->cBytes.copy_to_codehelper(ch);
            while(length < 4) { ch.add_byte(0); ++length; }
            return;
        }
        if(length == 4) {
            pB->cBytes.copy_to_codehelper(ch);
            return;
        }
        /* length > 4 */
        report_error(pB->get_line(), "Too many bytes in ID (4 expected)");
        pB->cBytes.copy_to_codehelper(ch, 4); return;
    }
    report_error(pId->get_line(), "Unrecognized field");
    ch.add_double(0ul);
}

/* }}}
 * {{{ SetTextNode, define-text, define-vehicle-name
 */
SetTextNode::SetTextNode(const string& sName, GenericNode *pGN)
        : Node(sName, pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pLanguage = 0; pFeature = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "language") {
            pLanguage = check_assign_field(pGN->get_line(), pLanguage, *iter);
            continue;
        }
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }

        StringNode *pS = dynamic_cast<StringNode*>(*iter);
        if(pS) {
            cStrings.push_back(pS); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SetTextNode::~SetTextNode()
{
    if(pLanguage) delete pLanguage;
    if(pFeature) delete pFeature;

    while(!cStrings.empty()) {
        delete cStrings.back();
        cStrings.pop_back();
    }
}

void SetTextNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    CodeHelper pseudo(iSpriteNumber, "*");

    pseudo.add_byte(4); // action 4

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // feature train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x10)) {
        report_error(get_line(), "feature field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    /* language */
    UINT32 lang_byte = decode_language(pLanguage, pMain);

    if(sName == string("define-text"))
        lang_byte |= 0x80;   // set high bit of language-id field

    pseudo.add_byte(lang_byte);

    /* number of entries */
    if(cStrings.size() == 0)
        report_error(get_line(), "Missing string entries");
    if(cStrings.size() > 255)
        report_error(get_line(), "Too many string entries");

    pseudo.add_byte(cStrings.size() & 0xff);

    /* first entry number */
    if(cStrings.size() > 0) {
        /* check numbers */
        int number = -1;
        list<StringNode*>::iterator iter;
        for(iter = cStrings.begin(); iter != cStrings.end(); ++iter) {
            FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>((*iter)->pNumber);
            if(pU) {
                if(number == -1)
                    number = pU->value;
                else {
                    if(number == (int)(pU->value)) {
                        ++number;
                    } else {
                        report_error(pU->get_line(),
                                     "Unexpected string number value");
                    }
                }
            } else {
                report_error((*iter)->get_line(),
                             "Missing or unrecognized number field");
            }
        }

        /* generate a word or a byte with first id */
        StringNode *pFirst = cStrings.front();
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pFirst->pNumber);
        if(pU) {
            if(sName == string("define-text")) {
                if(pU->value + cStrings.size() > 0xffff)
                    report_error(pU->get_line(), "Value too large");
                pseudo.add_word(pU->value & 0xffff);
            } else {
                if(pU->value + cStrings.size() > 0xff)
                    report_error(pU->get_line(), "Value too large");
                pseudo.add_byte(pU->value & 0xff);
            }
        } else {    // no number specified in first string node
            if(sName == string("define-text"))
                pseudo.add_word(0);
            else
                pseudo.add_byte(0);
        }

        /* generate the strings themselves */
        for(iter = cStrings.begin(); iter != cStrings.end(); ++iter) {
            FieldString *pS = dynamic_cast<FieldString*>((*iter)->pText);
            if(pS) {
                pS->cText.add_text(pseudo, 0); pseudo.add_byte(0);
            } else {
                report_error((*iter)->get_line(),
                             "Missing or unrecognized text field");
                pseudo.add_byte(0);
            }
        }
    }

    pseudo.output(fp);
}

/* }}}
 * {{{ StringNode
 */
StringNode::StringNode(GenericNode *pGN)
        : Node(string("string"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pNumber = 0; pText = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "number") {
            pNumber = check_assign_field(pGN->get_line(), pNumber, *iter);
            continue;
        }
        if(pF && pF->sName == "text") {
            pText = check_assign_field(pGN->get_line(), pText, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

StringNode::~StringNode()
{
    if(pNumber) delete pNumber;
    if(pText) delete pText;
}

/* }}}
 * {{{ RealSprite ('sprite')
 */
RealSprite::RealSprite(GenericNode *pGN)
        : Node(string("sprite"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFilename = 0; pXpos = 0; pYpos = 0; pCompression = 0;
    pXsize = 0; pYsize = 0; pXrel = 0; pYrel = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "filename") {
            pFilename = check_assign_field(pGN->get_line(), pFilename, *iter);
            continue;
        }
        if(pF && pF->sName == "x-pos") {
            pXpos = check_assign_field(pGN->get_line(), pXpos, *iter);
            continue;
        }
        if(pF && pF->sName == "y-pos") {
            pYpos = check_assign_field(pGN->get_line(), pYpos, *iter);
            continue;
        }
        if(pF && pF->sName == "x-size") {
            pXsize = check_assign_field(pGN->get_line(), pXsize, *iter);
            continue;
        }
        if(pF && pF->sName == "y-size") {
            pYsize = check_assign_field(pGN->get_line(), pYsize, *iter);
            continue;
        }
        if(pF && pF->sName == "x-rel") {
            pXrel = check_assign_field(pGN->get_line(), pXrel, *iter);
            continue;
        }
        if(pF && pF->sName == "y-rel") {
            pYrel = check_assign_field(pGN->get_line(), pYrel, *iter);
            continue;
        }
        if(pF && pF->sName == "compression") {
            pCompression = check_assign_field(pGN->get_line(),
                                              pCompression, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

RealSprite::~RealSprite()
{
    if(pFilename) delete pFilename;
    if(pXpos) delete pXpos;
    if(pYpos) delete pYpos;
    if(pCompression) delete pCompression;
    if(pXsize) delete pXsize;
    if(pYsize) delete pYsize;
    if(pXrel) delete pXrel;
    if(pYrel) delete pYrel;
}

void RealSprite::generate_nfo(int iSprite_num, MainNode *pMain, FILE *fp)
{
    string sFilename;
    UINT32 xpos, ypos, xsize, ysize;
    INT32 xrel, yrel;

    /* header + filename */
    FieldString *pFS = dynamic_cast<FieldString *>(pFilename);
    if(pFS) {
        sFilename = pFS->cText.convert_to_string();
    } else {
        report_error(get_line(), "Missing or unrecognized filename field");
        sFilename = "UNKNOWN";
    }
    CodeHelper pseudo(iSprite_num, sFilename);

    xpos = decode_unsigned(string("x-pos"), pXpos, ES_Double, 0);
    ypos = decode_unsigned(string("y-pos"), pYpos, ES_Double, 0);
    xsize = decode_unsigned(string("x-size"), pXsize, ES_Word, 1);
    ysize = decode_unsigned(string("y-size"), pYsize, ES_Byte, 1);
    xrel = decode_signed(string("x-rel"), pXrel, ES_Word, 0);
    yrel = decode_signed(string("y-rel"), pYrel, ES_Word, 0);

    if(xsize < 1 || xsize > 0xffff) {
        report_error(get_line(), "Field x-size is out of range");
        xsize = 1;
    }
    if(ysize < 1 || ysize > 0xff) {
        report_error(get_line(), "Field y-size is out of range");
        ysize = 1;
    }
    if(xsize * ysize > 0xffff) {
        report_error(get_line(), "Sprite too large (at most 65535 pixels)");
        xsize =1; ysize = 1;
    }
    if(xrel < -32768 || xrel > 32767) {
        report_error(get_line(), "Field x-rel is out of range");
        xrel = 0;
    }
    if(yrel < -32768 || yrel > 32767) {
        report_error(get_line(), "Field y-rel is out of range");
        yrel = 0;
    }

    pseudo.add_decimal(xpos);
    pseudo.add_decimal(ypos);

    TableEntry compr_entry;
    if(!bitset_idenvalue(string("compression"), pCompression,
                         string("compression_table"), ES_Byte, &compr_entry))
        compr_entry.set_value(0x1); // 0x01=color-0-transparent

    pseudo.add_byte(compr_entry.get_value());
    pseudo.add_decimal(ysize);
    pseudo.add_decimal(xsize);
    pseudo.add_decimal(xrel);
    pseudo.add_decimal(yrel);
    pseudo.output(fp);
}

/* }}}
 * {{{ SpriteSetsNode (+ CollectionNode)
 */
CollectionNode::CollectionNode(GenericNode *pGN)
        : Node(string("collection"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        RealSprite *pSpr = dynamic_cast<RealSprite*>(*iter);
        if(pSpr) {
            cSprites.push_back(pSpr); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

CollectionNode::~CollectionNode()
{
    while(!cSprites.empty()) {
        delete cSprites.back(); cSprites.pop_back();
    }
}

/* SpriteSetsNode */
SpriteSetsNode::SpriteSetsNode(GenericNode *pGN)
        : Node(string("sprite-sets"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        CollectionNode *pC = dynamic_cast<CollectionNode*>(*iter);
        if(pC) {
            cCollections.push_back(pC); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SpriteSetsNode::~SpriteSetsNode()
{
    if(pFeature) delete pFeature;
    while(!cCollections.empty()) {
        delete cCollections.back(); cCollections.pop_back();
    }
}

void SpriteSetsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    UINT32 num_sprites;

    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(1); // action 1

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x02))
        report_error(get_line(), "feature field has unsupported value");

    pseudo.add_byte(feature_entry.get_value());

    /* num-sets */
    if(cCollections.size() > 255)
        report_error(get_line(), "Too many sprite collections");

    pseudo.add_byte(cCollections.size() & 0xff);

    /* num-sprites in each set */
    if(cCollections.size() > 0) {
        num_sprites = cCollections.front()->cSprites.size();
    } else {
        num_sprites = 0;    // no collections available
    }

    if(num_sprites > 0xffff) {
        report_error(get_line(), "Too many sprites in each collection");
    }
    pseudo.add_extended(num_sprites & 0xffff);
    pseudo.output(fp);

    /* Action 1 pseudo-sprite done, now all sprites in each collection! */
    int sprite_num = iSpriteNumber + 1;
    list<CollectionNode*>::iterator coll_iter;
    for(coll_iter = cCollections.begin();
                                coll_iter != cCollections.end(); ++coll_iter) {
        /* check # sprites in this collection */
        if((*coll_iter)->cSprites.size() != num_sprites) {
            report_error((*coll_iter)->get_line(),
                                                "Incorrect number of sprites");
        }

        /* Generate real sprite data */
        list<RealSprite*>::iterator iter;
        for(iter = (*coll_iter)->cSprites.begin();
                                iter != (*coll_iter)->cSprites.end(); ++iter) {
            (*iter)->generate_nfo(sprite_num, pMain, fp);
            ++sprite_num;
        }
    }
}

int SpriteSetsNode::set_sprite_number(int num)
{
    UINT32 num_col, num_spr;

    iSpriteNumber = num;

    num_col = cCollections.size();
    if(num_col == 0)
        num_spr = 0;
    else
        num_spr = cCollections.front()->cSprites.size();

    return num + 1 + (num_col * num_spr);
}

CStringNumberMap SpriteSetsNode::getCollectionLabels() const
{
    CStringNumberMap cCollectionMap;
    int number = 0;
    CStringNumberMap::iterator iter;
    list<CollectionNode*>::const_iterator coll_iter;

    cCollectionMap.clear();
    for(coll_iter = cCollections.begin();
                                coll_iter != cCollections.end(); ++coll_iter) {
        string sL = (*coll_iter)->get_label();
        if(sL.size() > 0) {
            iter = cCollectionMap.find(sL);
            if(iter == cCollectionMap.end()) {
                cCollectionMap[sL] = number;
                ++number;
            } else {
                report_error((*coll_iter)->get_line(),
                                "Label of collection already used");
            }
        }
    }
    return cCollectionMap;
}

/* }}}
 * {{{ define-properties
 */

/** New value node */
NewValueNode::NewValueNode(GenericNode *pGN)
        : Node(string("new-value"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pProperty = 0; pValue = 0; pSize = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "property") {
            pProperty = check_assign_field(pGN->get_line(), pProperty, *iter);
            continue;
        }
        if(pF && pF->sName == "value") {
            pValue = check_assign_field(pGN->get_line(), pValue, *iter);
            continue;
        }
        if(pF && pF->sName == "size") {
            pSize = check_assign_field(pGN->get_line(), pSize, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

NewValueNode::~NewValueNode()
{
    if(pProperty) delete pProperty;
    if(pValue) delete pValue;
    if(pSize) delete pSize;
}

void NewValueNode::decode_fields(UINT32 feature)
{
    string sFeatureTable;
    UINT32 size;

    switch(feature) {
    case 0x00: // trains
        sFeatureTable = string("define-properties_train_table"); break;
    case 0x01: // road vehicles
        sFeatureTable = string("define-properties_road_table"); break;
    case 0x02: // ships
        sFeatureTable = string("define-properties_ship_table"); break;
    case 0x03: // planes
        sFeatureTable = string("define-properties_plane_table"); break;
    case 0x04: // stations
        sFeatureTable = string("define-properties_station_table"); break;
    case 0x05: // canals
        sFeatureTable = string("define-properties_canal_table"); break;
    case 0x06: // bridges
        sFeatureTable = string("define-properties_bridge_table"); break;
    case 0x07: // houses
        sFeatureTable = string("define-properties_house_table"); break;
    case 0x08: // global variables
        sFeatureTable = string("define-properties_global-var_table"); break;
    case 0x09: // industry tiles
        sFeatureTable = string("define-properties_industry-tile_table"); break;
    case 0x0A: // industries
        sFeatureTable = string("define-properties_industry_table"); break;
    case 0x0B: // cargos
        sFeatureTable = string("define-properties_cargo_table"); break;
    case 0x0C: // sound effects
        sFeatureTable = string("define-properties_sound_table"); break;
    case 0x0D: // airports
        sFeatureTable = string("define-properties_airport_table"); break;
    case 0x0F: // newobjects
        sFeatureTable = string("define-properties_object_table"); break;
    default:
        sFeatureTable = string(""); break;
    }

    TableEntry prop_entry;
    if(!unsigned_idenvalue(string("property"), pProperty,
                           sFeatureTable, ES_Byte, &prop_entry))
        prop_entry.set_value(0);    // default, not necessarilly useful

    cProperty = prop_entry.get_value();

    /* deduce size of the value
     */
    size = prop_entry.has_size() ? prop_entry.get_size() : 0;

    /* check for specified size */
    if(pSize) {
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pSize);
        if(pU) {
            if(size > 0 && size != pU->value) {
                report_error(pSize->get_line(),
                             string("Expected size of variable differs "
                                    "with specified size"));
            }
            size = pU->value;   // specified size takes precedence
        } else {
            report_error(pSize->get_line(), string("Unrecognized size field"));
        }
    }

    if(pValue) {
        FieldString *pS = dynamic_cast<FieldString*>(pValue);
        if(pS) {
            UINT32 cValSize = pS->cText.convert_to_string().size() + 1;
            if(size > 0 && size != cValSize) {
                report_error(pValue->get_line(),
                             string("Real value size differs from expected "
                                    "or specified size"));
            }
            size = cValSize;
        }
        FieldBytes *pB = dynamic_cast<FieldBytes*>(pValue);
        if(pB) {
            UINT32 cValSize = pB->cBytes.byte_length();
            if(size > 0 && size != cValSize) {
                report_error(pValue->get_line(),
                             string("Real value size differs from expected "
                                    "or specified size"));
            }
            size = cValSize;
        }
    } else {    // no value!! (maybe in the future correct for boolean fields?)
        report_error(get_line(), "Missing value field");
    }

    cPropSize = size;
}

/** IdValues node */
IdValuesNode::IdValuesNode(GenericNode *pGN)
        : Node(string("id-values"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pId = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "id") {
            pId = check_assign_field(pGN->get_line(), pId, *iter);
            continue;
        }
        NewValueNode *pNV = dynamic_cast<NewValueNode*>(*iter);
        if(pNV) {
            cNewValues.push_back(pNV); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

IdValuesNode::~IdValuesNode()
{
    if(pId) delete pId;
    while(!cNewValues.empty()) {
        delete cNewValues.back(); cNewValues.pop_back();
    }
}

void IdValuesNode::decode_fields(UINT32 feature)
{
    list<NewValueNode*>::iterator iter;

    cId = decode_unsigned(string("id"), pId, ES_Word, 0);
    if(cNewValues.size() == 0) {
        report_error(get_line(), string("Missing new-value nodes"));
        return;
    }
    for(iter = cNewValues.begin(); iter != cNewValues.end(); ++iter) {
        (*iter)->decode_fields(feature);
    }
}

/** Check that the node sets the same properties as the example list */
void IdValuesNode::check_newvalues(const list<NewValueNode*> *pNewValues)
{
    list<NewValueNode*>::const_iterator examp_iter, iter;

    for(examp_iter = pNewValues->begin();
                            examp_iter != pNewValues->end(); ++examp_iter) {
        bool found = false;
        for(iter = cNewValues.begin(); iter != cNewValues.end(); ++iter) {
            if((*examp_iter)->cProperty == (*iter)->cProperty) {
                found = true; break;
            }
        }
        if(found) {
            /* found, check sizes */
            if((*examp_iter)->cPropSize != (*iter)->cPropSize) {
                char buffer[256];

                sprintf(buffer, "Property size different from line %d",
                        (*examp_iter)->get_line());
                report_error(get_line(), string(buffer));
            }
                
        } else {
            char buffer[256];

            sprintf(buffer, "Cannot find property number %ld",
                    (*examp_iter)->cProperty);
            report_error(get_line(), string(buffer));
        }
    }
}

void IdValuesNode::generate_value(UINT32 cPropNum, CodeHelper& pseudo)
{
    list<NewValueNode*>::iterator iter;

    for(iter = cNewValues.begin(); iter != cNewValues.end(); ++iter) {
        if((*iter)->cProperty == cPropNum) {
            UINT32 cPropSize = (*iter)->cPropSize;
            FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>((*iter)->pValue);
            if(cPropSize > 0 && cPropSize <= 4 && pU) {
                switch(cPropSize) {
                case 4: pseudo.add_double(pU->value); break;
                case 3: pseudo.add_byte(pU->value & 0xff);
                        pseudo.add_byte(pU->value >> 8); break;
                case 2: pseudo.add_word(pU->value); break;
                case 1: pseudo.add_byte(pU->value); break;
                default: assert(0); // should never happen
                }
                return;
            }
            FieldString *pS = dynamic_cast<FieldString*>((*iter)->pValue);
            if(pS) {
                pS->cText.add_text(pseudo, 0); pseudo.add_byte(0);
                return;
            }
            FieldBytes *pB = dynamic_cast<FieldBytes*>((*iter)->pValue);
            if(pB) {
                pB->cBytes.copy_to_codehelper(pseudo);
                return;
            }

            report_error((*iter)->get_line(),
                         string("Unrecognized or missing value field"));
            return;
        }
    }
    /* silently not output anything (error should already been given) */
}

/** SetProperties node */
SetPropertiesNode::SetPropertiesNode(GenericNode *pGN)
        : Node(string("define-properties"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        IdValuesNode *pIV = dynamic_cast<IdValuesNode*>(*iter);
        if(pIV) {
            cIdValues.push_back(pIV); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SetPropertiesNode::~SetPropertiesNode()
{
    if(pFeature) delete pFeature;
    while(!cIdValues.empty()) {
        delete cIdValues.back(); cIdValues.pop_back();
    }
}

void SetPropertiesNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    list<IdValuesNode*>::iterator iter;

    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(0); // action 0

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // feature train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x01)) {
        report_error(get_line(), "feature field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    if(cIdValues.size() == 0) {
        report_error(get_line(), "Missing id-values nodes");
        /* output empty change set */
        pseudo.add_byte(0); pseudo.add_byte(0); pseudo.output(fp); return;
    }

    /* First, decode the fields to numeric values
     */
    for(iter = cIdValues.begin(); iter != cIdValues.end(); ++iter) {
        (*iter)->decode_fields(feature_entry.get_value());
    }

    /* Perform sanity checks
     */
    UINT32 expected_id = cIdValues.front()->cId;
    const list<NewValueNode*> *pNewValues = &(cIdValues.front()->cNewValues);
    for(iter = cIdValues.begin(); iter != cIdValues.end(); ++iter) {
        if((*iter)->cId != expected_id) {
            report_error((*iter)->get_line(),
                         string("Incorrect value of id field"));
        }
        ++expected_id;
        (*iter)->check_newvalues(pNewValues);
    }

    /* Generate code
     */
    pseudo.add_byte(pNewValues->size()); // num-props
    pseudo.add_byte(cIdValues.size());   // num-info
    pseudo.add_extended(cIdValues.front()->cId); // first ID

    list<NewValueNode*>::const_iterator examp_iter;
    for(examp_iter = pNewValues->begin();
                            examp_iter != pNewValues->end(); ++examp_iter) {
        pseudo.add_byte((*examp_iter)->cProperty);
        for(iter = cIdValues.begin(); iter != cIdValues.end(); ++iter) {
            (*iter)->generate_value((*examp_iter)->cProperty, pseudo);
        }
    }
    pseudo.output(fp);
}

/* }}}
 * {{{ VehicleGraphicsNode
 */
VehicleGraphicsNode::VehicleGraphicsNode(GenericNode *pGN)
    : Node(string("vehicle-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0; pCargoId = 0; pLoadTypes = 0; pMoveTypes = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }
        if(pF && pF->sName == "move-types") {
            pMoveTypes = check_assign_field(pGN->get_line(), pMoveTypes, *iter);
            continue;
        }
        if(pF && pF->sName == "load-types") {
            pLoadTypes = check_assign_field(pGN->get_line(), pLoadTypes, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

VehicleGraphicsNode::~VehicleGraphicsNode()
{
    if(pFeature) delete pFeature;
    if(pCargoId) delete pCargoId;
    if(pMoveTypes) delete pMoveTypes;
    if(pLoadTypes) delete pLoadTypes;
}

void VehicleGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    FieldLabels *pMoveLabels, *pLoadLabels;

    CodeHelper pseudo(iSpriteNumber, "*");

    pseudo.add_byte(2); // action2

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // default value: feature = train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x04)) {
        report_error(get_line(), "feature field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    /* add cargo-id */
    UINT32 cid = decode_unsigned(string("cargo-id"), pCargoId, ES_Byte, 0);
    pseudo.add_byte(cid);
    if(sLabel.size() > 0) {
        pMain->addGraphics(get_line(), sLabel, cid);
    }

    pMoveLabels = dynamic_cast<FieldLabels*>(pMoveTypes);
    if(!pMoveLabels) {
        report_error(get_line(), "Unknown or missing 'move-types' field");
    }

    pLoadLabels = dynamic_cast<FieldLabels*>(pLoadTypes);
    if(!pLoadLabels) {
        report_error(get_line(), "Unknown or missing 'load-types' field");
    }

    pseudo.add_byte(pMoveLabels ? pMoveLabels->cLabels.size() : 1);
    pseudo.add_byte(pLoadLabels ? pLoadLabels->cLabels.size() : 1);

    if(pMoveLabels) {
        add_collections(pMain, pseudo, pMoveLabels->cLabels);
    } else {
        pseudo.add_word(0);
    }

    if(pLoadLabels) {
        add_collections(pMain, pseudo, pLoadLabels->cLabels);
    } else {
        pseudo.add_word(0);
    }

    pseudo.output(fp);
}
        
/* }}}
 * {{{ building-graphics
 * {{{ BaseSprite
 */
BaseSprite::BaseSprite(const string& sName, const string& sLabel, int line)
        : Node(sName, sLabel, line)
{
    pUseSprite = 0; pDrawType = 0; pColorTranslation = 0;
    pSpriteFromAction1 = 0;
    pAlwaysDrawNormal = 0;
}

BaseSprite::~BaseSprite()
{
    if(pUseSprite) delete pUseSprite;
    if(pDrawType) delete pDrawType;
    if(pColorTranslation) delete pColorTranslation;
    if(pAlwaysDrawNormal) delete pAlwaysDrawNormal;
    if(pSpriteFromAction1) delete pSpriteFromAction1;
}

/** Compute the double representing the ground or building sprite data.
 * @param pMain: \c MainNode object
 * @param sColorTable: Name of the color-translation identifier table
 */
UINT32 BaseSprite::compute_sprite(MainNode *pMain, const string& sColorTable)
{
    UINT32 sprite_val = 0ul;

    bool bUsedLabel = false;

    /* bit 0-13: use-sprite */
    FieldIdenValue *pFI = dynamic_cast<FieldIdenValue*>(pUseSprite);
    if(pFI) {
        int iCollection = pMain->findCollection(pFI->sIdenValue);
        if(iCollection < 0) {
            report_error(pFI->get_line(), string("Collection \"") +
                         pFI->sIdenValue + string("\" not available here"));
            iCollection = 0;
        }
        sprite_val |= iCollection & 0x3fff;
        bUsedLabel = true;
    } else {
        FieldUnsigned *pU = dynamic_cast<FieldUnsigned*>(pUseSprite);
        if(pU) {
            sprite_val |= pU->value & 0x3fff;
            bUsedLabel = false;
        } else {
            report_error(get_line(),
                         string("Missing or unrecognized 'use-sprite' field"));
        }
    }

    /* bit 14-15: draw-type */
    TableEntry drawtype_entry;
    if(!unsigned_idenvalue(string("draw-type"), pDrawType,
                           string("draw-type_table"), ES_Byte, &drawtype_entry))
        drawtype_entry.set_value(0); // default 'normal'
    sprite_val |= ((drawtype_entry.get_value() & 3) << 14);

    /* bit 16-29: color-table */
    if(pColorTranslation) {
        TableEntry color_entry;
        if(!unsigned_idenvalue(string("color-translation"), pColorTranslation,
                               sColorTable, ES_Word, &color_entry))
            color_entry.set_value(0); // default no color translation
        sprite_val |= ((color_entry.get_value() & 0x3fff) << 16);
    }

    /* bit 30: always-draw-normal */
    if(pAlwaysDrawNormal) {
        FieldKey *pFK = dynamic_cast<FieldKey*>(pAlwaysDrawNormal);
        if(!pFK) {  // something else than a FieldKey object
            report_warning(pAlwaysDrawNormal->get_line(),
                           string("Value part of the always-draw-normal "
                                  "field ignored"));
        }
        sprite_val |= 1<<30;
    }

    /* bit 31: sprite-from-action-1 */
    if(bUsedLabel) {
        sprite_val |= 1<<31;
    } else {
        /* no label, is the sprite-from-action-1 flag set? */
        if(pSpriteFromAction1) {
            FieldKey *pFK = dynamic_cast<FieldKey*>(pSpriteFromAction1);
            if(!pFK) {  // something else than a FieldKey object
                report_warning(pSpriteFromAction1->get_line(),
                               string("Value part of the sprite-from-action-1 "
                                      "field ignored"));
            }
            sprite_val |= 1<<31;
        }
    }
    return sprite_val;
}

/* }}}
 * {{{ GroundNode
 */
GroundNode::GroundNode(GenericNode *pGN)
        : BaseSprite(string("ground"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    /* fields of BaseSprite already set to 0 */
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "use-sprite") {
            pUseSprite = check_assign_field(pGN->get_line(), pUseSprite, *iter);
            continue;
        }
        if(pF && pF->sName == "draw-type") {
            pDrawType = check_assign_field(pGN->get_line(), pDrawType, *iter);
            continue;
        }
        if(pF && pF->sName == "color-translation") {
            pColorTranslation = check_assign_field(pGN->get_line(),
                                                   pColorTranslation, *iter);
            continue;
        }
        if(pF && pF->sName == "sprite-from-action-1") {
            pSpriteFromAction1 = check_assign_field(pGN->get_line(),
                                                    pSpriteFromAction1, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

GroundNode::~GroundNode()
{
}

void GroundNode::add_sprite(MainNode *pMain, CodeHelper& ch)
{
    UINT32 sprite_val = BaseSprite::compute_sprite(pMain,
                                    string("ground_color-translation_table"));
    assert(!(sprite_val & (1<<30))); // bit 30 should be 0 */
    ch.add_double(sprite_val);
}

/* }}}
 * {{{ BuildingNode
 */
BuildingNode::BuildingNode(GenericNode *pGN)
        : BaseSprite(string("building"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    /* fields of BaseSprite already set to 0 */
    pXextent = 0; pYextent = 0; pZextent = 0;
    pXoffset = 0; pYoffset = 0; pZoffset = 0;
    pUseBoundingBoxBuilding = 0; pRelativeOffset = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "use-sprite") {
            pUseSprite = check_assign_field(pGN->get_line(), pUseSprite, *iter);
            continue;
        }
        if(pF && pF->sName == "draw-type") {
            pDrawType = check_assign_field(pGN->get_line(), pDrawType, *iter);
            continue;
        }
        if(pF && pF->sName == "color-translation") {
            pColorTranslation = check_assign_field(pGN->get_line(),
                                                   pColorTranslation, *iter);
            continue;
        }
        if(pF && pF->sName == "always-draw-normal") {
            pAlwaysDrawNormal = check_assign_field(pGN->get_line(),
                                                   pAlwaysDrawNormal, *iter);
            continue;
        }
        if(pF && pF->sName == "use-bounding-box-building") {
            pUseBoundingBoxBuilding = check_assign_field(pGN->get_line(),
                                                pUseBoundingBoxBuilding, *iter);
            continue;
        }
        if(pF && pF->sName == "relative-offset") {
            pRelativeOffset = check_assign_field(pGN->get_line(),
                                                 pRelativeOffset, *iter);
            continue;
        }
        if(pF && pF->sName == "sprite-from-action-1") {
            pSpriteFromAction1 = check_assign_field(pGN->get_line(),
                                                    pSpriteFromAction1, *iter);
            continue;
        }
        if(pF && pF->sName == "x-extent") {
            pXextent = check_assign_field(pGN->get_line(), pXextent, *iter);
            continue;
        }
        if(pF && pF->sName == "y-extent") {
            pYextent = check_assign_field(pGN->get_line(), pYextent, *iter);
            continue;
        }
        if(pF && pF->sName == "z-extent") {
            pZextent = check_assign_field(pGN->get_line(), pZextent, *iter);
            continue;
        }
        if(pF && pF->sName == "x-offset") {
            pXoffset = check_assign_field(pGN->get_line(), pXoffset, *iter);
            continue;
        }
        if(pF && pF->sName == "y-offset") {
            pYoffset = check_assign_field(pGN->get_line(), pYoffset, *iter);
            continue;
        }
        if(pF && pF->sName == "z-offset") {
            pZoffset = check_assign_field(pGN->get_line(), pZoffset, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

BuildingNode::~BuildingNode()
{
    if(pXextent) delete pXextent;
    if(pYextent) delete pYextent;
    if(pZextent) delete pZextent;
    if(pXoffset) delete pXoffset;
    if(pYoffset) delete pYoffset;
    if(pZoffset) delete pZoffset;
    if(pUseBoundingBoxBuilding) delete pUseBoundingBoxBuilding;
    if(pRelativeOffset) delete pRelativeOffset;
}

/** Compute the values of the x/y/z-extent and the x/y/z position fields.
 * Store the data in i[XYZ]extent, and i[XYZ]offset variables.
 * @param pPrev: Pointer to previous building node (or \c 0 if first node)
 * @return: Recommendation to use old or new-style building data (0=old, 1=new)
 */
int BuildingNode::compute_positions(const BuildingNode *pPrev)
{
    iXextent = decode_unsigned("x-extent", pXextent, ES_Byte, 0);
    iYextent = decode_unsigned("y-extent", pYextent, ES_Byte, 0);
    iZextent = decode_unsigned("z-extent", pZextent, ES_Byte, 0);
    iXoffset = decode_signed("x-offset", pXoffset, ES_Byte, 0);
    iYoffset = decode_signed("y-offset", pYoffset, ES_Byte, 0);
    iZoffset = decode_signed("z-offset", pZoffset, ES_Byte, 0);

    if(pRelativeOffset) {
        FieldKey *pFK = dynamic_cast<FieldKey*>(pRelativeOffset);
        if(!pFK) {
            report_warning(pRelativeOffset->get_line(),
                           string("Value part of the relative-offset "
                                  "field ignored"));
        }
        if(!pPrev) {
            report_error(pRelativeOffset->get_line(),
                         "Cannot use relative offset with first building");
        } else { // offsets are relative to previous, compute absolute values
            iXoffset += pPrev->iXoffset;
            iYoffset += pPrev->iYoffset;
            iZoffset += pPrev->iZoffset;
        }
    }

    if(pPrev) return 1; // the 2nd of further, always new-style
    if(iZoffset != 0) return 1; // old-style does not allow non-zero z-offset
    return 0;
}

void BuildingNode::write_oldstyle(MainNode *pMain, CodeHelper &pseudo)
{
    UINT32 sprite_val = BaseSprite::compute_sprite(pMain,
                                    string("building_color-translation_table"));
    pseudo.add_double(sprite_val);
    pseudo.add_byte(iXoffset & 0xff); pseudo.add_byte(iYoffset & 0xff);

    pseudo.add_byte(iXextent & 0xff); pseudo.add_byte(iYextent & 0xff);
    pseudo.add_byte(iZextent & 0xff);
}

void BuildingNode::write_newstyle(MainNode *pMain, CodeHelper &pseudo,
                                  const BuildingNode *pPrev)
{
    UINT32 sprite_val = BaseSprite::compute_sprite(pMain,
                                    string("building_color-translation_table"));
    pseudo.add_double(sprite_val);

    bool bBoundingboxBuilding; //< Use long bounding box new style building
    if(pPrev) {
        if(iXextent == pPrev->iXextent && iYextent == pPrev->iYextent &&
           iZextent == pPrev->iZextent && iZoffset == pPrev->iZoffset) {
            int dx = iXoffset - pPrev->iXoffset;
            int dy = iYoffset - pPrev->iYoffset;
            bBoundingboxBuilding = !(signed_in_range(dx, ES_Byte) &
                                     signed_in_range(dy, ES_Byte));
            // no bBoundingboxBuilding if both in signed byte range
        }
    } else {
        bBoundingboxBuilding = true;
    }

    if(pUseBoundingBoxBuilding) {
        FieldKey *pFK = dynamic_cast<FieldKey*>(pUseBoundingBoxBuilding);
        if(!pFK) {
            report_warning(pRelativeOffset->get_line(),
                           string("Value part of the use-bounding-box-building "
                                  "field ignored"));
        }
        bBoundingboxBuilding = true;
    }

    if(bBoundingboxBuilding) {
        pseudo.add_byte(iXoffset & 0xff); pseudo.add_byte(iYoffset & 0xff);
        pseudo.add_byte(iZoffset & 0xff);

        if((iZoffset & 0xff) == 0x80) {
            report_warning(get_line(), "Building may be interpreted as a "
                                       "relative offset building");
        }

        pseudo.add_byte(iXextent & 0xff); pseudo.add_byte(iYextent & 0xff);
        pseudo.add_byte(iZextent & 0xff);
    } else {
        // compact 'stacked' building
        int dx = iXoffset - pPrev->iXoffset;
        int dy = iYoffset - pPrev->iYoffset;
        pseudo.add_byte(dx & 0xff); pseudo.add_byte(dy & 0xff);
        pseudo.add_byte(0x80);
    }
}

/* }}}
 * {{{ BuildingGraphicsNode
 */
BuildingGraphicsNode::BuildingGraphicsNode(GenericNode *pGN)
        : Node(string("building-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0; pCargoId = 0; pGround = 0;
    cBuildings.clear();
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }
        GroundNode *pG = dynamic_cast<GroundNode*>(*iter);
        if(pG) {
            if(pGround) {
                char buffer[256];
                sprintf(buffer, "Overwriting field at line %d "
                                "by field at line %d",
                        pGround->get_line(), pG->get_line());
                report_warning(pG->get_line(), buffer);
                delete pGround;
            }
            pGround = pG; continue;
        }
        BuildingNode *pB = dynamic_cast<BuildingNode*>(*iter);
        if(pB) {
            cBuildings.push_back(pB); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

BuildingGraphicsNode::~BuildingGraphicsNode()
{
    if(pFeature) delete pFeature;
    if(pCargoId) delete pCargoId;
    if(pGround) delete pGround;
    while(!cBuildings.empty()) {
        delete cBuildings.back(); cBuildings.pop_back();
    }
}

void BuildingGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    list<BuildingNode*>::iterator iter;

    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(2); // action2

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // default value: feature = train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x20)) {
        report_error(get_line(), "features field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    /* add cargo-id */
    UINT32 cid = decode_unsigned(string("cargo-id"), pCargoId, ES_Byte, 0);
    pseudo.add_byte(cid);
    if(sLabel.size() > 0) {
        pMain->addGraphics(get_line(), sLabel, cid);
    }

    int style; /* old-style (0) / new-style (1) */
    size_t count = cBuildings.size();

    if(count == 0) style = 0;
    else if(count > 1) {
        const BuildingNode *pPrev = 0;
        for(iter = cBuildings.begin(); iter != cBuildings.end(); ++iter) {
            (*iter)->compute_positions(pPrev);
            pPrev = (*iter);
        }
        style = 1;  // more than 1 building
    } else {
        style = cBuildings.back()->compute_positions(0);
    }

    if(style == 0)
        pseudo.add_byte(0);
    else
        pseudo.add_byte(count);

    /* Write ground */
    if(pGround)
        pGround->add_sprite(pMain, pseudo);
    else
        pseudo.add_double(0ul);

    if(count == 0) {    // no building
        assert(style == 0); // we require old-style
        pseudo.add_double(0);
        pseudo.add_byte(0); pseudo.add_byte(0); pseudo.add_byte(0);
        pseudo.add_byte(0); pseudo.add_byte(0);
    } else {
        if(style == 0) { // old style
            assert(count == 1); // we require 1 building
            cBuildings.back()->write_oldstyle(pMain, pseudo);
        } else {    // new style
            const BuildingNode *pPrev = 0;
            for(iter = cBuildings.begin(); iter != cBuildings.end(); ++iter) {
                (*iter)->write_newstyle(pMain, pseudo, pPrev);
                pPrev = (*iter);
            }
        }
    }
    pseudo.output(fp);
}

/* }}}
 * }}}
 * {{{ StationGraphicsNode
 */
StationGraphicsNode::StationGraphicsNode(GenericNode *pGN)
        : Node(string("station-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pCargoId = 0; pLittleSets = 0; pLotsSets = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }
        if(pF && pF->sName == "little-sets") {
            pLittleSets = check_assign_field(pGN->get_line(),
                                             pLittleSets, *iter);
            continue;
        }
        if(pF && pF->sName == "lots-sets") {
            pLotsSets = check_assign_field(pGN->get_line(), pLotsSets, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

StationGraphicsNode::~StationGraphicsNode()
{
    if(pCargoId) delete pCargoId;
    if(pLittleSets) delete pLittleSets;
    if(pLotsSets) delete pLotsSets;
}

void StationGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    CodeHelper pseudo(iSpriteNumber, "*");

    pseudo.add_byte(2); // action2
    pseudo.add_byte(4); // feature = station

    /* add cargo-id */
    UINT32 cid = decode_unsigned(string("cargo-id"), pCargoId, ES_Byte, 0);
    pseudo.add_byte(cid);
    if(sLabel.size() > 0) {
        pMain->addGraphics(get_line(), sLabel, cid);
    }

    /* decide about little-sets */
    int iNumLittle;
    FieldLabels *pLittleLabels;
    if(pLittleSets) {
        pLittleLabels = dynamic_cast<FieldLabels*>(pLittleSets);
        if(pLittleLabels) {
            iNumLittle = pLittleLabels->cLabels.size();
        } else {
            report_error(pLittleSets->get_line(), "Expected list of labels");
            iNumLittle = 0;
        }
    } else {    // little-sets have been omitted from the command
        iNumLittle = 0; pLittleLabels = 0;
    }

    /* decide about lots-sets */
    int iNumLots;
    FieldLabels *pLotsLabels;
    if(pLotsSets) {
        pLotsLabels = dynamic_cast<FieldLabels*>(pLotsSets);
        if(pLotsLabels) {
            iNumLots = pLotsLabels->cLabels.size();
        } else {
            report_error(pLotsSets->get_line(), "Expected list of labels");
            iNumLots = 0;
        }
    } else {
        report_error(get_line(), "Missing lots-sets field");
        iNumLots = 0; pLotsLabels = 0;
    }

    pseudo.add_byte(iNumLittle);
    pseudo.add_byte(iNumLots);

    if(iNumLittle > 0) add_collections(pMain, pseudo, pLittleLabels->cLabels);
    if(iNumLots > 0) add_collections(pMain, pseudo, pLotsLabels->cLabels);

    pseudo.output(fp);
}

/* }}}
 * {{{ CargoGraphicsNode
 */
CargoGraphicsNode::CargoGraphicsNode(GenericNode *pGN)
        : Node(string("cargo-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pCargoId = 0; pCollection = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }
        if(pF && pF->sName == "collection") {
            pCollection = check_assign_field(pGN->get_line(),
                                             pCollection, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

CargoGraphicsNode::~CargoGraphicsNode()
{
    if(pCargoId) delete pCargoId;
    if(pCollection) delete pCollection;
}

void CargoGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    CodeHelper pseudo(iSpriteNumber, "*");

    pseudo.add_byte(2); // action2
    pseudo.add_byte(0x0b); // feature = cargo

    /* add cargo-id */
    UINT32 cid = decode_unsigned(string("cargo-id"), pCargoId, ES_Byte, 0);
    pseudo.add_byte(cid);
    if(sLabel.size() > 0) {
        pMain->addGraphics(get_line(), sLabel, cid);
    }

    pseudo.add_byte(1); pseudo.add_byte(0);

    FieldIdenValue *pFI = dynamic_cast<FieldIdenValue*>(pCollection);
    if(pFI) {
        int iCollection = pMain->findCollection(pFI->sIdenValue);
        if(iCollection < 0) {
            report_error(pFI->get_line(), string("Collection \"") +
                         pFI->sIdenValue + string("\" not available here"));
            iCollection = 0;
        }
        pseudo.add_word(iCollection);
    } else {
        report_error(get_line(), "Missing or unrecognized 'collection' field");
        pseudo.add_word(0); // kind of safe default sprite collection number
    }

    pseudo.output(fp);
}

/* }}}
 * {{{ IndustryCallbackNode
 */
IndustryCallbackNode::IndustryCallbackNode(GenericNode *pGN)
        : Node(string("industry-callback"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pId = 0; pVersion = 0;
    pIn1 = 0; pIn2 = 0; pIn3 = 0; pOut1 = 0; pOut2 = 0;
    pAgain = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "id") {
            pId = check_assign_field(pGN->get_line(), pId, *iter);
            continue;
        }
        if(pF && pF->sName == "version") {
            pVersion = check_assign_field(pGN->get_line(), pVersion, *iter);
            continue;
        }
        if(pF && pF->sName == "subtract-in-1") {
            pIn1 = check_assign_field(pGN->get_line(), pIn1, *iter);
            continue;
        }
        if(pF && pF->sName == "subtract-in-2") {
            pIn2 = check_assign_field(pGN->get_line(), pIn2, *iter);
            continue;
        }
        if(pF && pF->sName == "subtract-in-3") {
            pIn3 = check_assign_field(pGN->get_line(), pIn3, *iter);
            continue;
        }
        if(pF && pF->sName == "add-out-1") {
            pOut1 = check_assign_field(pGN->get_line(), pOut1, *iter);
            continue;
        }
        if(pF && pF->sName == "add-out-2") {
            pOut2 = check_assign_field(pGN->get_line(), pOut2, *iter);
            continue;
        }
        if(pF && pF->sName == "again") {
            pAgain = check_assign_field(pGN->get_line(), pAgain, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

IndustryCallbackNode::~IndustryCallbackNode()
{
    if(pId) delete pId;
    if(pVersion) delete pVersion;
    if(pIn1) delete pIn1;
    if(pIn2) delete pIn2;
    if(pIn3) delete pIn3;
    if(pOut1) delete pOut1;
    if(pOut2) delete pOut2;
    if(pAgain) delete pAgain;
}

/* }}}
 * {{{ select-graphics
 * {{{ CaseNode
 */
CaseNode::CaseNode(GenericNode *pGN)
        : Node(string("case"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pMin = 0; pMax = 0; pUseCargoId = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "minimal") {
            pMin = check_assign_field(pGN->get_line(), pMin, *iter);
            continue;
        }
        if(pF && pF->sName == "maximal") {
            pMax = check_assign_field(pGN->get_line(), pMax, *iter);
            continue;
        }
        if(pF && pF->sName == "use-cargo-id") {
            pUseCargoId = check_assign_field(pGN->get_line(),
                                             pUseCargoId, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

CaseNode::~CaseNode()
{
    if(pMin) delete pMin;
    if(pMax) delete pMax;
    if(pUseCargoId) delete pUseCargoId;
}

void CaseNode::write_nfo(MainNode *pMain, CodeHelper& ch, ESize es)
{
    UINT32 value;

    if(!add_cid(pMain, ch, pUseCargoId))
        report_error(get_line(), "Missing or unrecognized use-cargo-id field");

    value = decode_unsigned(string("minimal"), pMin, es, 0);
    write_sized_data(ch, value, es);
    value = decode_unsigned(string("maximal"), pMax, es, 0);
    write_sized_data(ch, value, es);
}

/* }}}
 * {{{ SelectComputeBase
 */
SelectComputeBase::SelectComputeBase(const string& sN, const string& sL, int l)
        : Node(sN, sL, l) { }
SelectComputeBase::~SelectComputeBase() { }

/* }}}
 * {{{ SelectVariable
 */
SelectVariable::SelectVariable(GenericNode *pGN)
        : SelectComputeBase(string("variable"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pVarName = 0; pShiftRight = 0; pAdd = 0; pAnd = 0;
    pDivide = 0; pModulo = 0; pParnum = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "name") {
            pVarName = check_assign_field(pGN->get_line(), pVarName, *iter);
            continue;
        }
        if(pF && pF->sName == "parnum") {
            pParnum = check_assign_field(pGN->get_line(), pParnum, *iter);
            continue;
        }
        if(pF && pF->sName == "shift-right") {
            pShiftRight = check_assign_field(pGN->get_line(),
                                             pShiftRight, *iter);
            continue;
        }
        if(pF && pF->sName == "add") {
            pAdd = check_assign_field(pGN->get_line(), pAdd, *iter);
            continue;
        }
        if(pF && pF->sName == "and") {
            pAnd = check_assign_field(pGN->get_line(), pAnd, *iter);
            continue;
        }
        if(pF && pF->sName == "divide") {
            pDivide = check_assign_field(pGN->get_line(), pDivide, *iter);
            continue;
        }
        if(pF && pF->sName == "modulo") {
            pModulo = check_assign_field(pGN->get_line(), pModulo, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SelectVariable::~SelectVariable()
{
    if(pVarName) delete pVarName;
    if(pShiftRight) delete pShiftRight;
    if(pAnd) delete pAnd;
    if(pAdd) delete pAdd;
    if(pDivide) delete pDivide;
    if(pModulo) delete pModulo;
    if(pParnum) delete pParnum;
}

/** Generate an NFO sequence of <variable> <var-adjust> for this SelectVariable
 * object.
 * @param pseudo: CodeHelper to write output to
 * @param sVarTable: Name of the table to use for decoding variable identifiers
 * @param es: Size of the data (from the <type> field)
 * @param bMoreVars: Set bit5 of <var-adjust> (more variables are coming)
 */
void SelectVariable::write_nfo(CodeHelper& pseudo, const string& sVarTable,
                               ESize es, bool bMoreVars)
{
    TableEntry var_entry;
    if(!unsigned_idenvalue(string("name"), pVarName, sVarTable, ES_Byte,
                           &var_entry))
        var_entry.set_value(0); // default value

    pseudo.add_byte(var_entry.get_value());

    /* check parnum presence w.r.t. type expectations of variable */
    if(var_entry.has_type()) {
        switch(var_entry.get_type()) {
        case 1: // no parnum expected
            if(pParnum) {
                report_error(pParnum->get_line(), "No parnum field expected");
            }
            break;
        case 2: // parnum expected
            if(!pParnum) {
                report_error(pParnum->get_line(), "Also expected parnum field");
            }
            break;
        default: assert(0); // should never happen
        }
    }

    /* add parnum parm if present */
    if(pParnum) {
        UINT32 cVal = decode_unsigned("parnum", pParnum, ES_Byte, 0);
        pseudo.add_byte(cVal);
    }

    /* shift-mask */
    UINT32 iShift = 0;
    if(pShiftRight) {
        iShift = decode_unsigned("shift-right", pShiftRight, ES_Byte, 0);
        if(iShift > 31) {
            report_error(pShiftRight->get_line(),
                         string("Shifting of at most 31 allowed"));
            iShift = 31;
        }
    }

    if(bMoreVars) {
        iShift |= 0x20;
    }

    if(pDivide) {
        if(pModulo) {   // report error when both divide and modulo
            report_error(get_line(), "Either divide or modulo field possible");
            delete pModulo; pModulo = 0;
        }
    }

    if(pDivide) iShift |= 0x40;
    if(pModulo) iShift |= 0x80;

    pseudo.add_byte(iShift);    // write shift mask

    /* and mask */
    UINT32 iAnd;
    switch(es) {    // default AND mask
    case ES_Byte: iAnd = 0xff; break;
    case ES_Word: iAnd = 0xffff; break;
    default: iAnd = 0xffffffff; break;
    };
    if(pAnd) {
        iAnd = decode_unsigned(string("and"), pAnd, es, iAnd);
    }

    write_sized_data(pseudo, iAnd, es); // write AND mask

    /* add */
    UINT32 iAdd;
    if(!pDivide && !pModulo && pAdd) {
        report_error(pAdd->get_line(), "Can only add together with either "
                                        "divide or modulo");
    }
    if(pAdd && (pDivide || pModulo)) {
        iAdd = decode_unsigned("add", pAdd, es, 0);
        write_sized_data(pseudo, iAdd, es); // write ADD value
    }

    /* divide */
    if(pDivide) {
        UINT32 cVal = decode_unsigned("divide", pDivide, es, 1);
        write_sized_data(pseudo, cVal, es); // write divide value
    }
    /* modulo */
    if(pDivide) {
        UINT32 cVal = decode_unsigned("modulo", pModulo, es, 1);
        write_sized_data(pseudo, cVal, es); // write modulo value
    }
}

/* }}}
 * {{{ SelectCompute
 */
SelectCompute::SelectCompute(GenericNode *pGN)
        : SelectComputeBase(string("compute"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pLeft = 0; pOperation = 0; pRight = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        SelectCompute *pC = dynamic_cast<SelectCompute*>(*iter);
        if(pC) {
            if(pLeft) {
                report_error(pC->get_line(), "Overwriting previous left side");
                delete pLeft;
            }
            pLeft = pC; continue;
        }
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "operation") {
            pOperation = check_assign_field(pGN->get_line(), pOperation, *iter);
            continue;
        }
        SelectVariable *pSV = dynamic_cast<SelectVariable*>(*iter);
        if(pSV && !pLeft) {
            pLeft = pSV; continue;
        }
        if(pSV) { // pLeft is taken
            if(pRight) {
                report_error(pSV->get_line(),
                             "Overwriting previous right side");
                delete pRight;
            }
            pRight = pSV; continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SelectCompute::~SelectCompute()
{
    if(pLeft) delete pLeft;
    if(pOperation) delete pOperation;
    if(pRight) delete pRight;
};

/** Generate an NFO sequence of <variable> <var-adjust> for this SelectVariable
 * object.
 * @param pseudo: CodeHelper to write output to
 * @param sVarTable: Name of the table to use for decoding variable identifiers
 * @param es: Size of the data (from the <type> field)
 * @param bMoreVars: Set bit5 of <var-adjust> (more variables are coming)
 */
void SelectCompute::write_nfo(CodeHelper& pseudo, const string& sVarTable,
                              ESize es, bool bMoreVars)
{
    if(pLeft) {
        pLeft->write_nfo(pseudo, sVarTable, es, true);
    } else {
        report_error(get_line(), "Missing left-hand side");
    }

    TableEntry op_entry;
    if(!unsigned_idenvalue(string("operation"), pOperation,
                           string("select-graphics_operations_table"), ES_Byte,
                           &op_entry))
        op_entry.set_value(1);  // default: add

    pseudo.add_byte(op_entry.get_value());

    if(pRight) {
        pRight->write_nfo(pseudo, sVarTable, es, bMoreVars);
    } else {
        report_error(get_line(), "Missing right-hand side");
    }
}

/* }}}
 * {{{ SelectGraphicsNode
 */
SelectGraphicsNode::SelectGraphicsNode(GenericNode *pGN)
        : Node(string("select-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0; pCargoId = 0; pType = 0; pDefaultCase = 0;
    pVarCompute = 0; cCases.clear();
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }
        if(pF && pF->sName == "type") {
            pType = check_assign_field(pGN->get_line(), pType, *iter);
            continue;
        }
        if(pF && pF->sName == "default-case") {
            pDefaultCase = check_assign_field(pGN->get_line(),
                                              pDefaultCase, *iter);
            continue;
        }
        SelectComputeBase *pS = dynamic_cast<SelectComputeBase*>(*iter);
        if(pS) {
            if(pVarCompute) {
                report_error(pS->get_line(),
                             "Overwriting previous variable/computation");
                delete pVarCompute;
            }
            pVarCompute = pS; continue;
        }
        CaseNode *pC = dynamic_cast<CaseNode*>(*iter);
        if(pC) {
            cCases.push_back(pC); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

SelectGraphicsNode::~SelectGraphicsNode()
{
    if(pFeature) delete pFeature;
    if(pCargoId) delete pCargoId;
    if(pType) delete pType;
    if(pVarCompute) delete pVarCompute;
    if(pDefaultCase) delete pDefaultCase;
    while(!cCases.empty()) { delete cCases.back(); cCases.pop_back(); }
}

void SelectGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(2); // action2 (variational action2)

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // default: feature train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x08)) {
        report_error(get_line(), "feature field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    /* add cargo-id */
    UINT32 cid = decode_unsigned(string("cargo-id"), pCargoId, ES_Byte, 0);
    pseudo.add_byte(cid);
    if(sLabel.size() > 0) {
        pMain->addGraphics(get_line(), sLabel, cid);
    }

    TableEntry type_entry;
    if(!unsigned_idenvalue(string("type"), pType,
                           string("select-graphics_type_table"), ES_Byte,
                           &type_entry))
        type_entry.set_value(0x81); // default: object-byte

    pseudo.add_byte(type_entry.get_value());

    /* decide the table name that should be used to decode the name of the
     * variables
     */
    UINT32 iType = type_entry.get_value();
    string sTableName = string(""); // default table: none
    switch(feature_entry.get_value()) {
    case 0x00:
    case 0x01:
    case 0x02:
    case 0x03:
        sTableName = string("select-graphics_vehicle_table"); break;
    case 0x04:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_station_table"); break;
        if(iType == 0x82 || iType == 0x86 || iType == 0x8a)
            sTableName = string("select-graphics_city_table"); break;
        break;  // otherwise, no table
    case 0x05:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_canal_table"); break;
        break;  // otherwise, no table
    case 0x07:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_house_table"); break;
        if(iType == 0x82 || iType == 0x86 || iType == 0x8a)
            sTableName = string("select-graphics_city_table"); break;
        break;  // otherwise, no table
    case 0x09:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_industry-tile_table"); break;
        if(iType == 0x82 || iType == 0x86 || iType == 0x8a)
            sTableName = string("select-graphics_industry_table"); break;
        break;  // otherwise, no table
    case 0x0a:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_industry_table"); break;
        if(iType == 0x82 || iType == 0x86 || iType == 0x8a)
            sTableName = string("select-graphics_city_table"); break;
        break;  // otherwise, no table
    case 0x0b:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_cargo_table"); break;
        break;  // otherwise, no table
    case 0x0e:
        if(iType == 0x81 || iType == 0x85 || iType == 0x89)
            sTableName = string("select-graphics_signal_table"); break;
        break;  // otherwise, no table
    }

    ESize eS = ES_Byte;
    if(iType == 0x81 || iType == 0x85 || iType == 0x89) eS = ES_Byte;
    if(iType == 0x82 || iType == 0x86 || iType == 0x8a) eS = ES_Word;

    if(pVarCompute) {
        pVarCompute->write_nfo(pseudo, sTableName, eS);
    } else {
        report_error(get_line(), string("Missing variable/compute node"));
    }

    /* n-cases */
    pseudo.add_byte(cCases.size());

    /* the cases */
    list<CaseNode*>::iterator iter;
    for(iter= cCases.begin(); iter != cCases.end(); ++iter) {
        (*iter)->write_nfo(pMain, pseudo, eS);
    }

    if(!add_cid(pMain, pseudo, pDefaultCase))
        report_error(get_line(), "Missing or unrecognized default-case field");

    pseudo.output(fp);
}

/* }}}
 * }}} end select-graphics
 * {{{ CargoNode, LinkGraphicsNode
 */
CargoNode::CargoNode(GenericNode *pGN)
        : Node(string("cargo"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pCargoType = 0; pCargoId = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "cargo-type") {
            pCargoType = check_assign_field(pGN->get_line(), pCargoType, *iter);
            continue;
        }
        if(pF && pF->sName == "cargo-id") {
            pCargoId = check_assign_field(pGN->get_line(), pCargoId, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

CargoNode::~CargoNode()
{
    if(pCargoType) delete pCargoType;
    if(pCargoId) delete pCargoId;
}

void CargoNode::generate_nfo(MainNode *pMain, CodeHelper& pseudo)
{
    /* cargo type */
    TableEntry cargotype_entry;
    if(!unsigned_idenvalue("cargo-type", pCargoType, "cargo-type_table",
                            ES_Byte, &cargotype_entry))
        cargotype_entry.set_value(0);   // possibly safe default

    pseudo.add_byte(cargotype_entry.get_value());

    if(!add_cid(pMain, pseudo, pCargoId))
        report_error(get_line(), "Missing or unrecognized cargo-id field");
}

/** Code for link-graphics node
*/
LinkGraphicsNode::LinkGraphicsNode(GenericNode *pGN)
        : Node(string("link-graphics"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    pFeature = 0; pDefaultCargoId = 0;
    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "feature") {
            pFeature = check_assign_field(pGN->get_line(), pFeature, *iter);
            continue;
        }
        if(pF && pF->sName == "livery-override") {
            pLiveryOverride = check_assign_field(pGN->get_line(),
                                                 pLiveryOverride, *iter);
            continue;
        }
        if(pF && pF->sName == "default-cargo-id") {
            pDefaultCargoId = check_assign_field(pGN->get_line(),
                                                 pDefaultCargoId, *iter);
            continue;
        }
        if(pF && pF->sName == "id") {
            cIds.push_back(pF); continue;
        }
        CargoNode *pC = dynamic_cast<CargoNode*>(*iter);
        if(pC) {
            cCargos.push_back(pC); continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

LinkGraphicsNode::~LinkGraphicsNode()
{
    if(pFeature) delete pFeature;
    if(pLiveryOverride) delete pLiveryOverride;
    if(pDefaultCargoId) delete pDefaultCargoId;
    while(!cIds.empty()) { delete cIds.back(); cIds.pop_back(); }
    while(!cCargos.empty()) { delete cCargos.back(); cCargos.pop_back(); }
}

void LinkGraphicsNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    UINT32 value;

    CodeHelper pseudo(iSpriteNumber, "*");
    pseudo.add_byte(3); // action3

    /* feature */
    TableEntry feature_entry;
    if(!unsigned_idenvalue(string("feature"), pFeature,
                           string("feature_table"), ES_Byte, &feature_entry))
        feature_entry.set_value(0); // default: feature train

    if(feature_entry.has_type() && !(feature_entry.get_type() & 0x08)) {
        report_error(get_line(), "feature field has unsupported value");
    }
    pseudo.add_byte(feature_entry.get_value());

    /* n-id byte */
    value = cIds.size();
    if(value > 0x7f) {
        report_error(get_line(), "Too many Id fields (at most 127 allowed)");
        value = 0x7f;
    }

    if(pLiveryOverride) {
        FieldKey *pFK = dynamic_cast<FieldKey*>(pLiveryOverride);
        if(!pFK) {  // It is something different than a FieldKey object
            report_warning(pLiveryOverride->get_line(),
                           string("Value part of the livery-override "
                                  "field ignored"));
        }
        value |= 0x80;
    }
    pseudo.add_byte(value);

    list<BaseField*>::iterator cid_iter;
    for(cid_iter = cIds.begin(); cid_iter != cIds.end(); ++cid_iter) {
        value = decode_unsigned("id", *cid_iter, ES_Byte, 0);
        pseudo.add_byte(value);
    }

    /* num-cid */
    value = cCargos.size();
    if(value > 0xff) {
        report_error(get_line(), "Too many cargo nodes");
        value = 0xff;
    }
    pseudo.add_byte(value);

    list<CargoNode*>::iterator cg_iter;
    for(cg_iter = cCargos.begin(); cg_iter != cCargos.end(); ++cg_iter) {
        (*cg_iter)->generate_nfo(pMain, pseudo);
    }

    /* default-cargo */
    if(!add_cid(pMain, pseudo, pDefaultCargoId))
        report_error(get_line(),
                     string("Missing or unrecognized default-cargo-id field"));

    pseudo.output(fp);
}

/* }}}
 * {{{ GrlBytesNode
 */
GrlBytesNode::GrlBytesNode(GenericNode *pGN)
    : Node(string("grl-bytes"), pGN->sLabel, pGN->get_line())
{
    list<BaseField*>::iterator iter;

    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Field *pF = dynamic_cast<Field*>(*iter);
        if(pF && pF->sName == "data") {
            pData = check_assign_field(pGN->get_line(), pData, *iter);
            continue;
        }

        report_error((*iter)->get_line(), "Unrecognized field");
        delete *iter;
    }
    pGN->cFields.clear();
}

GrlBytesNode::~GrlBytesNode()
{
    if(pData) delete pData;
}

void GrlBytesNode::generate_nfo(MainNode *pMain, FILE *fp)
{
    FieldBytes *pB = dynamic_cast<FieldBytes*>(pData);
    if(pB) {
        CodeHelper pseudo(iSpriteNumber, "*");
        pB->cBytes.copy_to_codehelper(pseudo);
        pseudo.output(fp);
    } else {
        report_error(get_line(), "Missing or unrecognized 'data' field");
    }
}

/* }}}

 * {{{ decode_node() function
 */
BaseField *decode_node(GenericNode *gn)
{
    assert(gn);

    if(gn->sName == "cond" || gn->sName == "init-cond") {
        BaseField *pBF = new CondJumpNode(gn->sName, gn);
        delete gn; return pBF;
    }
    if(gn->sName == "message") {
        BaseField *pBF = new MessageNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "grl-id") {
        BaseField *pBF = new GrlIdNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "string") {
        BaseField *pBF = new StringNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "define-text") {
        BaseField *pBF = new SetTextNode(gn->sName, gn);
        delete gn; return pBF;
    }
    if(gn->sName == "define-vehicle-name") {
        BaseField *pBF = new SetTextNode(gn->sName, gn);
        delete gn; return pBF;
    }
    if(gn->sName == "sprite") {
        BaseField *pBF = new RealSprite(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "collection") {
        BaseField *pBF = new CollectionNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "sprite-sets") {
        BaseField *pBF = new SpriteSetsNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "new-value") {
        BaseField *pBF = new NewValueNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "id-values") {
        BaseField *pBF = new IdValuesNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "define-properties") {
        BaseField *pBF = new SetPropertiesNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "vehicle-graphics") {
        BaseField *pBF = new VehicleGraphicsNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "station-graphics") {
        BaseField *pBF = new StationGraphicsNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "cargo-graphics") {
        BaseField *pBF = new CargoGraphicsNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "grl-bytes") {
        BaseField *pBF = new GrlBytesNode(gn);
        delete gn; return pBF;
    }

    /* building-graphics */
    if(gn->sName == "ground") {
        BaseField *pBF = new GroundNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "building") {
        BaseField *pBF = new BuildingNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "building-graphics") {
        BaseField *pBF = new BuildingGraphicsNode(gn);
        delete gn; return pBF;
    }

    /* select-graphics */
    if(gn->sName == "select-graphics") {
        BaseField *pBF = new SelectGraphicsNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "case") {
        BaseField *pBF = new CaseNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "variable") {
        BaseField *pBF = new SelectVariable(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "computation") {
        BaseField *pBF = new SelectCompute(gn);
        delete gn; return pBF;
    }

    /* link-graphics */
    if(gn->sName == "cargo") {
        BaseField *pBF = new CargoNode(gn);
        delete gn; return pBF;
    }
    if(gn->sName == "link-graphics") {
        BaseField *pBF = new LinkGraphicsNode(gn);
        delete gn; return pBF;
    }


    string sError = "Node \"" + gn->sName + "\" not recognized";
    report_error(gn->get_line(), sError);

    delete gn;
    return 0;
}

/* }}}

 * {{{ MainNode */

MainNode::MainNode(GenericNode *pGN)
{
    list<BaseField*>::iterator iter;

    for(iter = pGN->cFields.begin(); iter != pGN->cFields.end(); ++iter) {
        Node *pN = dynamic_cast<Node*>(*iter);
        assert(pN); // dynamic_cast should always succeed
        cNodes.push_back(pN);
    }
    pGN->cFields.clear();
}
MainNode::~MainNode()
{
    while(!cNodes.empty()) {
        delete cNodes.back();
        cNodes.pop_back();
    }
}

/** Check the GRL sprites, and generate NFO output to \a fp
 */
void MainNode::generate_nfo(FILE *fp)
{
    int iSprnum;
    list<Node*>::iterator iter;
    map<string, Node*>::iterator map_iter;

    // Check static semantics
    cCollectionLabels.clear();
    cGraphicsLabels.clear();
    find_grlid();

    // Allocate sprite numbers
    iSprnum = 1;
    for(iter = cNodes.begin(); iter != cNodes.end(); ++iter) {
        iSprnum = (*iter)->set_sprite_number(iSprnum);
    }
    // Collect labels to jump to
    cJumpDest.clear();
    for(iter = cNodes.begin(); iter != cNodes.end(); ++iter) {
        string sL = (*iter)->get_label();
        if(sL.size() > 0) {
            map_iter = cJumpDest.find(sL);
            if(map_iter != cJumpDest.end()) {
                report_error((*iter)->get_line(), "Label already used");
            } else {
                cJumpDest[sL] = *iter;
            }
        }
    }

    /* Generate the NFO file
     * While generating the etries, collect the sprite collection labels of the
     * last 'sprite-sets' (action1) entry
     */
    generate_sprite0(iSprnum - 1, fp);
    for(iter = cNodes.begin(); iter != cNodes.end(); ++iter) {
        (*iter)->generate_nfo(this, fp);    // Generate code for sprite (*iter)
        SpriteSetsNode *pSprSets = dynamic_cast<SpriteSetsNode*>(*iter);
        if(pSprSets) {
            cCollectionLabels = pSprSets->getCollectionLabels();
        }
    }
}

/** Find the index in the last sprite-sets command of the collection indicated
 * by the label.
 * @param sLabel: Label of the collecion we are looking for
 * @return: The index of the looked for collection, or \c -1 if not available
 */
int MainNode::findCollection(const string& sLabel)
{
    CStringNumberMap::const_iterator iter;

    iter = cCollectionLabels.find(sLabel);
    if(iter == cCollectionLabels.end()) return -1;
    return (*iter).second;
}

void MainNode::addGraphics(int line, const string& sLabel, UINT32 value)
{
    CStringNumberMap::iterator iter;

    /* Delete label with the same value from map (if existing) */
    for(iter = cGraphicsLabels.begin(); iter != cGraphicsLabels.end(); ++iter) {
        if((*iter).second == value) {
            cGraphicsLabels.erase(iter); break;
        }
    }
    iter = cGraphicsLabels.find(sLabel);
    if(iter != cGraphicsLabels.end()) {
        report_error(line, string("Label '") + sLabel +
                           string("' already defined"));
    }

    cGraphicsLabels[sLabel] = value;
}

/** Find the Id of the *-graphics node indicated by the label.
 * @param sLabel: Label of the collecion we are looking for
 * @return: The index of the looked for collection, or \c -1 if not available
 */
int MainNode::findGraphics(const string& sLabel)
{
    CStringNumberMap::iterator iter;

    iter = cGraphicsLabels.find(sLabel);
    if(iter != cGraphicsLabels.end()) {
        return (*iter).second;
    }
    return -1;
}

/** Find grl-id node.
 * Found node in put in MainNode::pGrlid
 */
void MainNode::find_grlid()
{
    list<Node*>::iterator iter;

    pGrlid = 0;
    for(iter = cNodes.begin(); iter != cNodes.end(); ++iter) {
        GrlIdNode *pGid = dynamic_cast<GrlIdNode*>(*iter);
        if(pGid) {
            if(pGrlid) {
                report_error(pGid->get_line(),
                             "Too many grl-id nodes (only one allowed)");
            } else {
                pGrlid = pGid;
            }
        }
    }

    if(!pGrlid) {
        report_error(0, "Missing grl-id node");
    }
}

Node *MainNode::get_destination(const string& sD)
{
    map<string, Node*>::iterator map_iter;

    map_iter = cJumpDest.find(sD);
    if(map_iter == cJumpDest.end())
        return 0;

    return (*map_iter).second;
}

/** Generate sprite number 0, containing the number of sprites in the file
 */
void MainNode::generate_sprite0(UINT32 num_spr, FILE *fp)
{
    CodeHelper pseudo(0);
    pseudo.add_double(num_spr);
    pseudo.output(fp);
}

int MainNode::get_grlid_version()
{
    if(!pGrlid) {
        return 6;   // no grl-id in the file available, return a default
    }
    return pGrlid->get_version();
}

/* }}}
 * {{{ decode_mainnode() function */

/** Decode function for making the top-level node \c MainNode
 */
MainNode *decode_mainnode(GenericNode *gn)
{
    MainNode *pMain;

    assert(gn->sName == "main-node");

    pMain = new MainNode(gn);
    delete gn; return pMain;
}

/* }}} */

