//
// $Id: node.h 2 2007-12-30 09:48:00Z hat $
//
#ifndef H_NODE
#define H_NODE

#include "types.h"
#include "generic-ast.h"

enum ESize
{
    ES_Byte = 1,
    ES_Word = 2,
    ES_Double = 4
};

typedef map<string, UINT32> CStringNumberMap;

/* {{{ Nodes
 */
struct DataValue;
class MainNode;
class TableEntry;

/* {{{ Base class for nodes
 */
class Node: public BaseField
{
public:
    Node(const string& sName, const string& sLabel, int line);
    ~Node();

    virtual string get_name(); //< Name of the node
    string get_label() const;

    int get_sprite_number();
    virtual int set_sprite_number(int sprnum);


    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    int iSpriteNumber;
    const string sName,     //< Name of the node
                 sLabel;    //< Label-prefix of the node (or \c "")

    bool unsigned_idenvalue(const string& sFieldName, BaseField *pField,
                            const string& sTableName, ESize eSize,
                            TableEntry* pData);

    UINT32 decode_unsigned(const string& sFieldName,
                    BaseField *pField, ESize eSize, UINT32 default_val);
    INT32 decode_signed(const string& sFieldName,
                    BaseField *pField, ESize eSize, INT32 default_val);

    bool bitset_idenvalue(const string& sFieldName, BaseField *pField,
                          const string& sTableName, ESize eSize,
                          TableEntry* pData);

    void generate_string(int node_line, const string& sName,
                         BaseField *pBF, CodeHelper& ch);

    UINT32 decode_language(BaseField *pLanguage, MainNode *pMain);
};

/* }}}
 * {{{ Conditional jump node
 */
class CondJumpNode: public Node
{
public:
    CondJumpNode(const string&, GenericNode*);
    ~CondJumpNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    bool init_cond;

    BaseField *pVariable;
    BaseField *pTest;
    BaseField *pValue;
    BaseField *pVarsize;
    BaseField *pDestination;
};

/* }}}
 * {{{ Report error message
 */
class MessageNode: public Node
{
public:
    MessageNode(GenericNode *);
    ~MessageNode();

    virtual void generate_nfo(MainNode*, FILE*);
protected:
    BaseField *pSeverity;
    BaseField *pText;
    BaseField *pData;
    BaseField *pLanguage;
    BaseField *pParnum1;
    BaseField *pParnum2;
};

/* }}}
 * {{{ Set ID of the file
 */
class GrlIdNode: public Node
{
public:
    GrlIdNode(GenericNode *);
    ~GrlIdNode();

    int get_version();
    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pVersion;
    BaseField *pId;
    BaseField *pName;
    BaseField *pDescription;
    int version;

    void generate_id(CodeHelper&);
};

/* }}}
 * {{{ Set-text
 */
class StringNode: public Node
{
public:
    StringNode(GenericNode *);
    ~StringNode();
//protected:
    BaseField *pNumber;
    BaseField *pText;
};

class SetTextNode: public Node
{
public:
    SetTextNode(const string& sN, GenericNode *);
    ~SetTextNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pLanguage;
    BaseField *pFeature;
    list<StringNode*> cStrings;
};

/* }}}
 * {{{ sprite-sets
 */
class RealSprite: public Node
{
public:
    RealSprite(GenericNode *);
    ~RealSprite();

    virtual void generate_nfo(int sprnum, MainNode *pMain, FILE *fp);
//protected:
    BaseField *pFilename;
    BaseField *pXpos;
    BaseField *pYpos;
    BaseField *pCompression;
    BaseField *pXsize;
    BaseField *pYsize;
    BaseField *pXrel;
    BaseField *pYrel;
};

class CollectionNode: public Node
{
public:
    CollectionNode(GenericNode *);
    ~CollectionNode();
//protected:
    list<RealSprite*> cSprites;
};

class SpriteSetsNode: public Node
{
public:
    SpriteSetsNode(GenericNode *);
    ~SpriteSetsNode();

    virtual int set_sprite_number(int sprnum);
    virtual void generate_nfo(MainNode *pMain, FILE *fp);
    CStringNumberMap getCollectionLabels() const;
protected:
    BaseField *pFeature;
    list<CollectionNode*> cCollections;
};

BaseField *decode_node(GenericNode *);

/* }}}
 * {{{ define-properties node
 */
class NewValueNode: public Node
{
public:
    NewValueNode(GenericNode *);
    ~NewValueNode();

    void decode_fields(UINT32 feature);
//protected:
    BaseField *pProperty;
    BaseField *pValue;
    BaseField *pSize;
    UINT32 cProperty;   //< Property as numeric value
    UINT32 cPropSize;   //< Deduced property-value size
};

class IdValuesNode: public Node
{
public:
    IdValuesNode(GenericNode *);
    ~IdValuesNode();

    void decode_fields(UINT32 feature);
    void check_newvalues(const list<NewValueNode*> *pNewValues);
    void generate_value(UINT32 cPropNum, CodeHelper& pseudo);
//protected:
    BaseField *pId;
    list<NewValueNode*> cNewValues;
    UINT32 cId;
};

class SetPropertiesNode: public Node
{
public:
    SetPropertiesNode(GenericNode *);
    ~SetPropertiesNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pFeature;
    list<IdValuesNode*> cIdValues;
};

/* }}}
 * {{{ VehicleGraphicsNode
 */
class VehicleGraphicsNode: public Node
{
public:
    VehicleGraphicsNode(GenericNode *);
    ~VehicleGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pFeature;
    BaseField *pCargoId;
    BaseField *pMoveTypes;
    BaseField *pLoadTypes;
};

/* }}}
 * {{{ BuildingGraphicsNode, BuildingNode, GroundNode
 */
class BaseSprite: public Node
{
public:
    BaseSprite(const string& sName, const string& sLabel, int line);
    ~BaseSprite();
protected:
    BaseField *pUseSprite;
    BaseField *pDrawType;
    BaseField *pColorTranslation;
    BaseField *pAlwaysDrawNormal;
    BaseField *pSpriteFromAction1;

    UINT32 compute_sprite(MainNode *pMain, const string& sColorTable);
};

class GroundNode: public BaseSprite
{
public:
    GroundNode(GenericNode *);
    ~GroundNode();

    void add_sprite(MainNode *pMain, CodeHelper& ch);
};

class BuildingNode: public BaseSprite
{
public:
    BuildingNode(GenericNode *);
    ~BuildingNode();

    int compute_positions(const BuildingNode *pPrev);
    void write_oldstyle(MainNode *pMain, CodeHelper &pseudo);
    void write_newstyle(MainNode *pM, CodeHelper &ch, const BuildingNode *prv);
protected:
    BaseField *pXextent, *pYextent, *pZextent;
    BaseField *pXoffset, *pYoffset, *pZoffset;
    BaseField *pUseBoundingBoxBuilding, *pRelativeOffset;

    UINT32 iXextent, iYextent, iZextent; //< Absolute coordinates bounding box
    INT32 iXoffset, iYoffset, iZoffset;  //< Absolute offsets
};

class BuildingGraphicsNode: public Node
{
public:
    BuildingGraphicsNode(GenericNode *);
    ~BuildingGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pFeature;
    BaseField *pCargoId;
    GroundNode *pGround;
    list<BuildingNode*> cBuildings;
};

/* }}}
 * {{{ StationGraphicsNode
 */
class StationGraphicsNode: public Node
{
public:
    StationGraphicsNode(GenericNode *pGN);
    ~StationGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pCargoId;
    BaseField *pLittleSets;
    BaseField *pLotsSets;
};

/* }}}
 * {{{ CargoGraphicsNode
 */
class CargoGraphicsNode: public Node
{
public:
    CargoGraphicsNode(GenericNode *);
    ~CargoGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pCargoId;
    BaseField *pCollection;
};

/* }}}
 * {{{ IndustryCallbackNode
 */
class IndustryCallbackNode: public Node
{
public:
    IndustryCallbackNode(GenericNode *);
    ~IndustryCallbackNode();
protected:
    BaseField *pId;
    BaseField *pVersion;
    BaseField *pIn1;
    BaseField *pIn2;
    BaseField *pIn3;
    BaseField *pOut1;
    BaseField *pOut2;
    BaseField *pAgain;
};

/* }}}
 * {{{ SelectGraphicsNode
 */
class SelectComputeBase: public Node
{
public:
    SelectComputeBase(const string& sName, const string& sLabel, int line);
    ~SelectComputeBase();

    virtual void write_nfo(CodeHelper& pseudo, const string& sVarTable,
                           ESize es, bool bMoreVars=false) = 0;
};

class SelectVariable: public SelectComputeBase
{
public:
    SelectVariable(GenericNode *pGN);
    ~SelectVariable();

    virtual void write_nfo(CodeHelper& pseudo, const string& sVarTable,
                           ESize es, bool bMoreVars=false);
protected:
    BaseField *pVarName;
    BaseField *pShiftRight;
    BaseField *pAnd;
    BaseField *pAdd;
    BaseField *pDivide;
    BaseField *pModulo;
    BaseField *pParnum;
};
class SelectCompute: public SelectComputeBase
{
public:
    SelectCompute(GenericNode *pGN);
    ~SelectCompute();

    virtual void write_nfo(CodeHelper& pseudo, const string& sVarTable,
                           ESize es, bool bMoreVars=false);
protected:
    SelectComputeBase *pLeft;
    BaseField *pOperation;
    SelectVariable *pRight;
};

class CaseNode: public Node
{
public:
    CaseNode(GenericNode *);
    ~CaseNode();
    
    void write_nfo(MainNode *pMain, CodeHelper &ch, ESize es);
protected:
    BaseField *pMin;
    BaseField *pMax;
    BaseField *pUseCargoId;
};

class SelectGraphicsNode: public Node
{
public:
    SelectGraphicsNode(GenericNode *pGN);
    ~SelectGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pFeature;
    BaseField *pCargoId;
    BaseField *pType;
    SelectComputeBase *pVarCompute;
    list<CaseNode*> cCases;
    BaseField *pDefaultCase;
};

/* }}}
 * {{{ LinkGraphicsNode
 */
class CargoNode: public Node
{
public:
    CargoNode(GenericNode *);
    ~CargoNode();

    void generate_nfo(MainNode *pmain, CodeHelper& pseudo);
protected:
    BaseField *pCargoType;
    BaseField *pCargoId;
};

class LinkGraphicsNode: public Node
{
public:
    LinkGraphicsNode(GenericNode *);
    ~LinkGraphicsNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pFeature;
    BaseField *pDefaultCargoId;
    BaseField *pLiveryOverride;
    list<BaseField *>cIds;
    list<CargoNode*> cCargos;
};

/* }}}
 * {{{ GrlBytesNode
 */
class GrlBytesNode: public Node
{
public:
    GrlBytesNode(GenericNode *);
    ~GrlBytesNode();

    virtual void generate_nfo(MainNode *pMain, FILE *fp);
protected:
    BaseField *pData;
};

/* }}}
 * }}}
 */
/* {{{ MainNode
 */

/** Node representing the GRL file
 */
class MainNode
{
public:
    MainNode(GenericNode *);
    ~MainNode();

    void generate_nfo(FILE *fp);
    Node *get_destination(const string& sD);
    int get_grlid_version();
    int findCollection(const string& sLabel);

    void addGraphics(int line, const string& sLabel, UINT32 value);
    int findGraphics(const string& sLabel);
protected:
    list<Node*> cNodes;         //< List nodes/sprites in the file

    map<string, Node*> cJumpDest; //< Labels that may be used as for jumping
    CStringNumberMap cCollectionLabels; //< Labels pointing to sprite collection
    CStringNumberMap cGraphicsLabels; //< Labels + cargo-id values
    GrlIdNode *pGrlid;  //< Pointer to the GrlId node

    void find_grlid();
    void generate_sprite0(UINT32 num_spr, FILE *fp);
};

MainNode *decode_mainnode(GenericNode *);

/* }}}
 */
#endif

/*

 * {{{ xxxx
 */
/* }}}

 */
