/*
 * $Id: textstring.cc 2 2007-12-30 09:48:00Z hat $
 */
#include <cassert>
#include "textstring.h"
#include "errors.h"

TextFragment::TextFragment(int ftype, const std::string& st)
		: frag_type(ftype), sText(st) { }
TextFragment::TextFragment(const TextFragment& tf)
		: frag_type(tf.frag_type), sText(tf.sText) { }
TextFragment& TextFragment::operator=(const TextFragment& tf)
{
	if(this != &tf) {
		frag_type = tf.frag_type;
		sText = tf.sText;
	}
	return *this;
}

TextFragment::~TextFragment() { }

bool operator==(const TextFragment& tf1, const TextFragment& tf2)
{
    if(tf1.frag_type != tf2.frag_type) return false;
    if(tf1.sText != tf2.sText) return false;
    return true;
}

/*
 * Text strings
 */

/** Constructor, return empty string
 */
TextString::TextString() { clear(); }

/** Constructor, load contents from \a sText.
 */
TextString::TextString(const string& sText)
{
    string::const_iterator iter1, iter2;

    clear();
    iter1 = sText.begin();
    while(iter1 != sText.end()) {
        if(*iter1 != '%') {
            iter2 = iter1;
            do { ++iter2; } while(iter2 != sText.end() && *iter2 != '%');
            add_character(string(iter1, iter2));
            iter1 = iter2;
        } else {
            /* *iter1 == '%' */
            iter2 = iter1; ++iter2;
            assert(iter2 != sText.end()); assert(*iter2 == '{');
            ++iter2;
            assert(iter2 != sText.end()); assert(*iter2 != '}');
            if(*iter2 == '0') { // hexadecimal number
                ++iter2; assert(iter2 != sText.end());
                assert(*iter2 == 'x' || *iter2 == 'X');
                do {
                    ++iter2;
                } while(iter2 != sText.end() &&
                        ((*iter2 >= '0' && *iter2 <= '9') ||
                         (*iter2 >= 'A' && *iter2 <= 'F') ||
                         (*iter2 >= 'a' && *iter2 <= 'f')));
                assert(iter2 != sText.end()); assert(*iter2 == '}');
                ++iter2;
                add_escape_byte(string(iter1, iter2));
                iter1 = iter2;
            } else {
                do { ++iter2; } while(iter2 != sText.end() && *iter2 != '}');
                assert(iter2 != sText.end());
                ++iter2;
                add_escape_text(string(iter1, iter2));
                iter1 = iter2;
            }
        }
    }
}
TextString::TextString(const TextString& ts): sFragments(ts.sFragments) { }
TextString& TextString::operator=(const TextString& ts)
{
	if(this != & ts) {
		sFragments = ts.sFragments;
	}
	return *this;
}
TextString::~TextString() { }

void TextString::clear() { sFragments.clear(); }

void TextString::add_character(const string& sText)
{
    if(sFragments.size() > 0) {
        TextFragment& cFrag = sFragments.back();
        if(cFrag.frag_type == FRAGTYPE_CHARS) {
            cFrag.sText += sText; return;
        }
    }
    sFragments.push_back(TextFragment(FRAGTYPE_CHARS, sText));
}

void TextString::add_escape_text(const string& sText)
{
    assert(sText.size() > 3 + 1);
    string sText2 = sText.substr(2, sText.size() - 3);
    sFragments.push_back(TextFragment(FRAGTYPE_ESC_TEXT, sText2));
}

void TextString::add_escape_byte(const string& sText)
{
    // 3 for "%{" + "}", and 3 for "0[xX][0-9a-fA-F]"
    assert(sText.size() >= 3 + 3);
    string sText2 = sText.substr(2, sText.size() - 3);
    sFragments.push_back(TextFragment(FRAGTYPE_ESC_BYTE, sText2));
}

bool TextString::empty() const { return sFragments.empty(); }

string TextString::convert_to_string()
{
	std::list<TextFragment>::iterator iter;

	std::string s = "";

	for(iter = sFragments.begin();iter != sFragments.end(); ++iter) {
		switch((*iter).frag_type) {
		case FRAGTYPE_CHARS:
			s += (*iter).sText; break;
		case FRAGTYPE_ESC_TEXT:
                    if((*iter).sText == "percent") {
                        s += string("%"); break;
                    }
                    if((*iter).sText == "quote") {
                        s += string("\""); break;
                    }
                    assert(0); // no other escape sequences exist
                    break;
		case FRAGTYPE_ESC_BYTE:
		{
			int val = strtol((*iter).sText.c_str(), 0, 16);
			char buf[2] = { val & 0xff, 0 };
			s += std::string(buf); break;
		}
		default: assert(0); break;
		}
	}
	return s;
}

void TextString::add_text(CodeHelper& ch, const DataValue *pDValues)
{
	static const DataValue dummy_datavalues[] = { {0, 0xffff} };
	if(!pDValues) {
		pDValues = dummy_datavalues;
	}

	std::list<TextFragment>::iterator iter;
	for(iter = sFragments.begin();iter != sFragments.end(); ++iter) {
		switch((*iter).frag_type) {
		case FRAGTYPE_CHARS:
		{
			const char *pText = (*iter).sText.c_str();
			while(*pText) {
				ch.add_byte(*pText);
				++pText;
			}
			break;
		}
		case FRAGTYPE_ESC_TEXT:
		{
			UINT32 val = decode_esc_text((*iter).sText, pDValues);
			ch.add_byte(val); break;
		}
		case FRAGTYPE_ESC_BYTE:
		{
			UINT32 val = strtol((*iter).sText.c_str(), 0, 16);
			ch.add_byte(val & 0xff); break;
		}
		default: assert(0); break;
		}
	}
}

/** Find an escaped text. Return the index or \c -1 if it cannot be found
 * @param sValue: Escaped text-value to find
 * @param iStart: Start index
 */
int TextString::find_esc_text(const string& sValue, int iStart)
{
    int idx = 0;
    std::list<TextFragment>::iterator iter;

    for(iter = sFragments.begin();iter != sFragments.end(); ++iter) {
        if(idx >= iStart && (*iter).frag_type == FRAGTYPE_ESC_TEXT &&
           (*iter).sText == sValue)
            return idx;
        ++idx;
    }
    return -1;
}

/** Decode the string of an \c FRAGTYPE_ESC_TEXT text fragment to a byte
 * @param sText: Text of the escaped text fragment
 * @param pDValues: List of text/value pairs of user-defined escape sequences.
 *	Default value is a value larger than \c 0xff
 * @return: A byte to replace the escaped text with
 **/
UINT32 TextString::decode_esc_text(const std::string& sText,
				   const DataValue *pDValues)
{
	/* Check user-supplied escape sequences */
	if(pDValues) {
                while(pDValues->text) {
                    if(pDValues->text == sText) {
                        return pDValues->value & 0xff;
                    }
                    ++pDValues;
                }
	}

	/* Check default escape sequences */
	if(sText == "percent") return '%';
	if(sText == "quote")   return '\"';

	/* Cannot decode, return an error */
	string sError = string("Cannot decode '%{") +
			       sText + string("}'");
	report_error(0, sError.c_str());

	return 0xAE;	/* return dummy value */
}

/** concatenation function of two TextStrings
 *
 * General strategy: If one of them is empty, return the other.
 * Otherwise, make a copy of the first, and concatenate each element of the
 * second after it.
*/
TextString operator+(const TextString& t1, const TextString& t2)
{
	if(t1.empty()) return t2;
	if(t2.empty()) return t1;

	TextString txt_str(t1);

	bool first = true;
	std::list<TextFragment>::const_iterator iter;
	for(iter = t2.sFragments.begin(); iter != t2.sFragments.end(); ++iter) {
		/* The first time, the last fragment of the first string
		 * may be characters, and the frist fragment of the second
		 * string may also be characters.
		 * In that case, both fragments should be merged.
		 */
		if(first) {
			first = false;
			if((*iter).frag_type == FRAGTYPE_CHARS) {
				TextFragment &final = txt_str.sFragments.back();
				if(final.frag_type == FRAGTYPE_CHARS) {
					final.sText += (*iter).sText;
				} else {
					txt_str.sFragments.push_back(*iter);
				}
			} else {
				txt_str.sFragments.push_back(*iter);
			}
		} else {
			txt_str.sFragments.push_back(*iter);
		}
	}
	return txt_str;
}

std::ostream& operator<<(std::ostream& os, const TextString& ts)
{
	std::list<TextFragment>::const_iterator iter;
	os << "\"";
	for(iter = ts.sFragments.begin(); iter != ts.sFragments.end(); ++iter) {
		if((*iter).frag_type == FRAGTYPE_CHARS)
			os << (*iter).sText;
		else
			os << "%{" << (*iter).sText << "}";
	}
	os << "\""; return os;
}

bool operator==(const TextString& ts1, const TextString& ts2)
{
    if(ts1.sFragments.size() != ts2.sFragments.size())
        return false;

    std::list<TextFragment>::const_iterator iter1, iter2;
    iter1 = ts1.sFragments.begin(); iter2 = ts2.sFragments.begin();
    while (iter1 != ts1.sFragments.end()) {
        if(!(*iter1 ==*iter2))
            return false;
        ++iter1; ++iter2;
    }
    return true;
}

